# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from .const import MRT_BGP4MP_ENTRY_ANY
from .const import MRT_BGP4MP_ENTRY_AS4_ANY
from .const import MRT_BGP4MP_ENTRY_ADDPATH_ANY
from .const import MRT_BGP4MP_STATE_CHANGE
from .const import MRT_BGP4MP_STATE_CHANGE_AS4
from ..bgp.msg import unpack_mrt_bgp_msg
from ...util import CACHE_TS
from ...const import IPV4
from ...const import IPV6
from ...const import IPV4_STR
from ...const import IPV6_STR
from ...const import AF_INET
from ...const import AF_INET6
from ...const import AFI_IPV4
from ...const import AFI_IPV6
from ...const import STRUCT_2B
from ...const import STRUCT_4B
from ...const import STRUCT_8B8B
from ...const import DATETIME_FORMAT_USEC
from ...const import DATETIME_FORMAT_MIN
from ...const import struct_unpack
from ...const import socket_inet_ntop
from ...const import datetime_utcfromtimestamp
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_AS
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_IP
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES_HUMAN
from ....model.const import FTL_ATTR_BGP_STATE_CHANGE_STATE_TO_STR
from ....model.const import FTL_ATTR_BGP_STATS_BGP_MESSAGE_TYPE_TO_STR
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_TYPE
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL
from ....model.record import FTL_RECORD_BGP_STATE_CHANGE
from ....model.record import FTL_RECORD_BGP_STATS
from ....model.error import FtlMrtDataError


def unpack_mrt_entry_bgp4mp(caches, stats_record, bgp_error, state_change_records, route_records, keep_alive_records,
                            route_refresh_records, notification_records, open_records, entry_bytes, mrt_subtype,
                            sequence, ts):
    """ Parse MRT BGP4MP entry.
    """
    # Check for AS4 support
    aslen, asbytelen = 2, STRUCT_2B
    if mrt_subtype in MRT_BGP4MP_ENTRY_AS4_ANY:
        aslen, asbytelen = 4, STRUCT_4B

    # Check MRT entry subtype
    elif mrt_subtype not in MRT_BGP4MP_ENTRY_ANY:
        yield from bgp_error(FtlMrtDataError(f'Unknown MRT subtype ({mrt_subtype}) for BGP4MP entry',
                                             reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_TYPE, data=entry_bytes))
        return

    ##################
    # BGP4MP PARSING #
    ##################

    # ----------------------------------------------------
    # [RFC8050] 3. MRT Subtypes for Types BGP4MP/BGP4MP_ET
    # ----------------------------------------------------
    # This document defines the following new subtypes:
    #  o  BGP4MP_MESSAGE_ADDPATH
    #  o  BGP4MP_MESSAGE_AS4_ADDPATH
    #  o  BGP4MP_MESSAGE_LOCAL_ADDPATH
    #  o  BGP4MP_MESSAGE_AS4_LOCAL_ADDPATH

    # Check for ADD-PATH entry (RFC7911)
    addpath = mrt_subtype in MRT_BGP4MP_ENTRY_ADDPATH_ANY

    # Prepare byte offset
    offset = 0

    # ----------------------------------------------------------------------------
    # [RFC6396] 4.4.1. BGP4MP_STATE_CHANGE Subtype / 4.4.2. BGP4MP_MESSAGE Subtype
    # ----------------------------------------------------------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |         Peer AS Number        |        Local AS Number        |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |        Interface Index        |        Address Family         |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                      Peer IP Address (variable)               |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                      Local IP Address (variable)              |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # ------------------------------------------------------------------------------------
    # [RFC6396] 4.4.3. BGP4MP_MESSAGE_AS4 Subtype / 4.4.4. BGP4MP_STATE_CHANGE_AS4 Subtype
    # ------------------------------------------------------------------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                         Peer AS Number                        |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                         Local AS Number                       |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |        Interface Index        |        Address Family         |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                      Peer IP Address (variable)               |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                      Local IP Address (variable)              |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Parse peer AS
    peer_as = struct_unpack(asbytelen, entry_bytes[offset:offset + aslen])[0]
    offset += aslen

    # Skip local AS
    offset += aslen

    # Skip interface index
    offset += 2

    # Parse AFI value
    afi = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
    offset += 2

    # Check AFI value
    if afi != AFI_IPV4 and afi != AFI_IPV6:  # pylint: disable=consider-using-in
        yield from bgp_error(FtlMrtDataError(f'Unknown AFI value ({afi}) in BGP4MP entry',
                                             reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=entry_bytes))
        return

    # Parse peer IP
    afinet, iplen = (AF_INET6, 16) if afi == AFI_IPV6 else (AF_INET, 4)
    peer_ip = entry_bytes[offset:offset + iplen]
    offset += iplen

    # Skip local IP
    offset += iplen

    #######################
    # BGP4MP STATE_CHANGE #
    #######################

    # ------------------------------------------------------------------------------------
    # [RFC6396] 4.4.1. BGP4MP_STATE_CHANGE Subtype / 4.4.4 BGP4MP_STATE_CHANGE_AS4 Subtype
    # ------------------------------------------------------------------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |            Old State          |          New State            |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Handle state change entry
    # pylint: disable-next=consider-using-in
    if mrt_subtype == MRT_BGP4MP_STATE_CHANGE or mrt_subtype == MRT_BGP4MP_STATE_CHANGE_AS4:

        # Access state change template
        state_change_init, state_change_emit, _ = state_change_records

        # Initialize state change record
        state_change_record = list(state_change_init)

        # Add timestamp to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP >= 0:
            ts_state_change = ts

            # Cache date string for minute-based timestamp
            if FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP_HUMAN:
                if caches:
                    ts_cache = caches[CACHE_TS]
                    ts_min, ts_sec = divmod(ts_state_change, 60)
                    ts_cached = ts_cache.get(ts_min, None)
                    if ts_cached is None:
                        ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                        ts_cache[ts_min] = ts_cached

                    # Add seconds and microseconds
                    ts_state_change = f'{ts_cached}:{ts_sec:09.6f}'

                # Do not use cache
                else:
                    ts_state_change = datetime_utcfromtimestamp(ts_state_change).strftime(DATETIME_FORMAT_USEC)

            # Add timestamp
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_TIMESTAMP] = ts_state_change

        # Add peer protocol to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL >= 0:
            if FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL_HUMAN:
                state_change_record[FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL] = (IPV6_STR if afinet == AF_INET6
                                                                                else IPV4_STR)
            else:
                state_change_record[FTL_ATTR_BGP_STATE_CHANGE_PEER_PROTOCOL] = IPV6 if afinet == AF_INET6 else IPV4

        # Add peer AS to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_PEER_AS >= 0:
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_PEER_AS] = peer_as

        # Add peer IP to state change record
        if FTL_ATTR_BGP_STATE_CHANGE_PEER_IP >= 0:
            peer_ip_state_change = peer_ip
            if FTL_ATTR_BGP_STATE_CHANGE_PEER_IP_HUMAN:
                peer_ip_state_change = socket_inet_ntop(afinet, peer_ip_state_change)
            elif afinet == AF_INET6:
                net, host = struct_unpack(STRUCT_8B8B, peer_ip_state_change)
                peer_ip_state_change = (net << 64) + host
            else:
                peer_ip_state_change = struct_unpack(STRUCT_4B, peer_ip_state_change)[0]
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_PEER_IP] = peer_ip_state_change

        # Parse old state
        if FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE >= 0:
            old_state = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
            if FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE_HUMAN:
                old_state = FTL_ATTR_BGP_STATE_CHANGE_STATE_TO_STR.get(old_state, str(old_state))
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_OLD_STATE] = old_state
        offset += 2

        # Parse new state
        if FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE >= 0:
            new_state = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
            if FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE_HUMAN:
                new_state = FTL_ATTR_BGP_STATE_CHANGE_STATE_TO_STR.get(new_state, str(new_state))
            state_change_record[FTL_ATTR_BGP_STATE_CHANGE_NEW_STATE] = new_state
        offset += 2

        # Yield final state change record
        if FTL_RECORD_BGP_STATE_CHANGE:
            yield state_change_emit(state_change_record)
        return

    ##################
    # BGP4MP MESSAGE #
    ##################

    # ---------------------------------------------------------------------------
    # [RFC6396] 4.4.2. BGP4MP_MESSAGE Subtype / 4.4.3. BGP4MP_MESSAGE_AS4 Subtype
    # ---------------------------------------------------------------------------
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                    BGP Message... (variable)
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # -----------------------------------
    # [RFC4271] 4.1 Message Header Format
    # -----------------------------------
    # 0                   1                   2                   3
    # 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                                                               |
    # +                                                               +
    # |                                                               |
    # +                                                               +
    # |                           Marker                              |
    # +                                                               +
    # |                                                               |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |          Length               |      Type     |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Skip marker
    offset += 16

    # Skip message length
    offset += 2

    # Parse message type
    mtype = entry_bytes[offset]
    offset += 1

    # Update stats record
    if FTL_RECORD_BGP_STATS:

        # Add BGP message type
        if FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES >= 0:
            bgptype = mtype
            if FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES_HUMAN:
                bgptype = FTL_ATTR_BGP_STATS_BGP_MESSAGE_TYPE_TO_STR.get(mtype, mtype)
            bgptype = str(bgptype)
            stats_record_mrt_bgp_msg = stats_record[FTL_ATTR_BGP_STATS_MRT_BGP_MESSAGE_TYPES]
            stats_record_mrt_bgp_msg[bgptype] = stats_record_mrt_bgp_msg.get(bgptype, 0) + 1

    # Parse message
    msg_bytes = entry_bytes[offset:]
    yield from unpack_mrt_bgp_msg(caches, stats_record, bgp_error, route_records, keep_alive_records,
                                  route_refresh_records, notification_records, open_records, msg_bytes, mtype, sequence,
                                  ts, peer_as, peer_ip, afinet, aslen=aslen, addpath=addpath)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

__author__ = 'Johann SCHLAMP'
__copyright__ = 'Copyright (C) 2014-2025 Leitwert GmbH'
__license__ = 'MIT license'

__all__ = [
    'MrtParser',
    'LglParser',
    'BgpParser',
    'FtlError',
    'FtlFileError',
    'FtlFormatError',
    'FtlDataError'
]

# Local imports
from .version import __version__
from .parser import FtlParser
from .model.error import FtlError
from .model.error import FtlFileError
from .model.error import FtlFormatError
from .model.error import FtlDataError
from .data.mrt.unpack import unpack_mrt_data
from .data.lgl.unpack import unpack_lgl_data


@FtlParser(unpack_mrt_data)
def MrtParser():
    """ Parse BGP data in MRT format.
    """


@FtlParser(unpack_lgl_data)
def LglParser():
    """ Parse BGP data in looking glass format.
    """


@FtlParser(unpack_mrt_data, unpack_lgl_data)
def BgpParser():
    """ Parse BGP data in MRT or looking glass format.
    """



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
# flake8: noqa
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import functools
import json
import re
import sys
import threading
from collections import namedtuple

# Local imports
from .const import FTL_ATTR
from .const import FTL_RECORD
from .error import FtlError

# Module patching
FTL_PATCH_THREAD_LOCK = threading.Lock()
FTL_PATCH_LOCKED = '__patchlock__'
FTL_PATCH_HASH = '__patchhash__'

# Bit shift for human-readable attributes
FTL_ATTR_SHIFT_HUMAN = 32


def generate_spec_attributes(cls, spec):
    """ Generate generic attributes specification instance.
    """
    # Prepare spec fields
    fields = dict(**{attr.name: attr for attr in spec})

    # Return spec instance
    return namedtuple(cls, fields.keys())(**fields)


def generate_spec_record(cls, spec, human=True, typed=None):
    """ Generate generic record specification instance.
    """
    # Extract name from class
    parts = re.sub('([A-Z]+)', r' \1', cls).split()
    parts = parts[2:] if len(parts) > 3 else parts[1:]
    spectype, name = parts[0].lower(), '_'.join(p.lower() for p in parts[1:])
    argname = '_'.join((spectype, name) if typed is not None else (name, spectype))
    istyped = typed is not None

    # Prepare base record spec
    # [Format] spec := (field, value, default)
    base_spec = (
        ('internal',   set(),      None),
        ('spec',       spectype,   None),
        ('name',       name,       None),
        ('argname',    argname,    None),
        ('typed',      istyped,    None),
        ('ALL',        0,          None),
        ('NONE',       0,          None),
        ('DEFAULT',    0,          None),
    )

    # Prepare human-readable record spec
    # [Format] spec := (field, value, default)
    if human is True:
        base_spec += (
            ('human', None, None),
        )

    # Prepare typed record spec
    # [Format] spec := (field, value, default)
    if istyped is True:
        base_spec += (
            ('type', typed, True),
        )

    # Prepare record spec data
    fields, values, internal = dict(), set(), {field for field, _, _ in base_spec}

    # Iterate record spec fields
    for field, value, default in base_spec + spec:

        # Check values for duplicates
        if default is not None:
            if field in fields:
                field_error = 'Internal' if field in internal else 'Existing'
                raise FtlError(f'{field_error} field "{field}" redefined in model specification {cls}')
            if value in values:
                field_error = value.bit_length() - 1
                raise FtlError(f'Existing value "1 << {field_error}" redefined in model specification {cls}')
            values.add(value)

            # Add value to all/default-type
            fields['ALL'] |= value
            if default is True:
                fields['DEFAULT'] |= value

        # Update value fields
        fields[field] = value

        # Update internal fields
        if default is None:
            fields['internal'].add(field)

    # Add human-readable record spec
    if human is True:
        fields_human = {field: value << FTL_ATTR_SHIFT_HUMAN for field, value in fields.items()
                        if field not in fields['internal'] or field in {'ALL', 'NONE', 'DEFAULT'}}
        fields['human'] = namedtuple(f'{cls}Human', fields_human.keys())(**fields_human)

    # Return record spec instance
    return namedtuple(cls, fields.keys())(**fields)


def apply_spec_records(cls, named_records, serialize, raise_on_errors, handle_error, records, attributes, spec_records,
                       spec_attributes):
    """ Setup generic record templates based on given record/attributes spec.
    """
    # Prepare attribute-to-index mapping
    records_field_to_value = dict()
    attrs_field_to_idx = dict()
    attrs_field_to_human = dict()

    # Prepare template fields
    template_fields = dict()
    for record_name, record_value in records._asdict().items():
        if record_name not in records.internal:
            record_attrs = getattr(attributes, record_name)
            spec_attrs = getattr(spec_attributes, record_name)

            # Skip unrequested records
            if not spec_records & record_value:
                template_fields[record_name] = (tuple(), tuple, None)
                continue

            # Prepare record type and template
            record_attrs_field_to_idx = attrs_field_to_idx.setdefault(record_name, dict())
            record_attrs_field_to_human = attrs_field_to_human.setdefault(record_name, dict())
            record_type = record_name if (spec_attrs >> FTL_ATTR_SHIFT_HUMAN) & record_attrs.type else record_value
            records_field_to_value[record_name] = record_type
            record_init = list()

            # Map attributes to tuple fields/indexes
            attr_fields, attr_index = list(), 0
            for attr_field, attr_value in record_attrs._asdict().items():
                if attr_field not in record_attrs.internal:
                    if spec_attrs & attr_value or (spec_attrs >> FTL_ATTR_SHIFT_HUMAN) & attr_value:

                        # Prepare initial record value (None or auto-typed)
                        record_init_value = None
                        if record_attrs.typed is True:
                            if record_attrs.type == attr_value:
                                record_init_value = record_type

                        # Update record template and index mapping
                        record_init.append(record_init_value)
                        record_attrs_field_to_idx[attr_field] = attr_index
                        record_attrs_field_to_human[attr_field] = (spec_attrs >> FTL_ATTR_SHIFT_HUMAN) & attr_value != 0

                        # Update record fields
                        attr_fields.append(attr_field)
                        attr_index += 1

            # Prepare emit method
            record = ''.join((part.capitalize() for part in record_name.split('_')))
            record_emit = namedtuple(f'{cls}{record}Record', attr_fields)._make if named_records else tuple
            if serialize:
                if named_records:
                    # pylint: disable-next=unnecessary-lambda-assignment
                    record_emit = lambda record, attr_fields=attr_fields: json.dumps(dict(zip(attr_fields, record)))
                else:
                    # pylint: disable-next=unnecessary-lambda-assignment
                    record_emit = lambda record: ','.join(str(e)[1:-1].replace(',', '').replace('\'', '')
                                                          if isinstance(e, tuple) else str(e)
                                                          if e is not None else '' for e in record)

            # Register initialized record template
            template_fields[record_name] = (tuple(record_init), record_emit, record_type)

    # Access stats record template
    stats_init, stats_emit, stats_record = tuple(), tuple, None
    try:
        stats_init, stats_emit, _ = template_fields[attributes.stats.name]
        stats_record = list(stats_init)
    except AttributeError:
        pass

    # Access error record template
    error_init, error_emit, error_type = tuple(), tuple, None
    try:
        error_init, error_emit, error_type = template_fields[attributes.error.name]
    except AttributeError:
        pass

    # Finalize record templates
    for record_name, (record_init, record_emit, record_type) in template_fields.items():

        # Parametrize record error function
        record_error = functools.partial(handle_error, raise_on_errors, error_init, error_emit, record_type, stats_emit,
                                         stats_record)

        # Add record error function to record template
        template_fields[record_name] = (record_init, record_emit, record_error)

    # Finalize record templates and generic record error function
    record_templates = namedtuple(f'FtlRecordTemplates{cls}', template_fields.keys())(**template_fields)
    record_error = functools.partial(handle_error, raise_on_errors, error_init, error_emit, error_type, stats_emit,
                                     stats_record)

    # Return finalized record spec
    return (record_templates, record_error, records_field_to_value, attrs_field_to_idx, attrs_field_to_human, stats_emit,
            stats_record)


def patch_spec_records(module, records, attributes, records_field_to_value, attrs_field_to_idx, attrs_field_to_human):
    """ Patch modules with given record/attributes spec.
    """
    # Prepare record flags
    const_to_record_flag = dict()
    records_name = records.name.upper()
    for record_field in records._fields:
        if record_field not in records.internal:
            record_flag = record_field in records_field_to_value
            const_to_record_flag['_'.join((FTL_RECORD, records_name, record_field.upper()))] = record_flag

    # Prepare attribute indexes
    const_to_attrs_idx = dict()
    for attrs in attributes:
        attrs_name = attrs.argname.upper()
        field_to_idx = attrs_field_to_idx.get(attrs.name, dict())
        for attr_field in attrs._fields:
            if attr_field not in attrs.internal:
                attr_idx = field_to_idx.get(attr_field, -1)
                const_to_attrs_idx['_'.join((FTL_ATTR, attrs_name, attr_field.upper()))] = attr_idx

    # Prepare human-readable attribute flags
    const_to_human_attrs_flag = dict()
    for attrs in attributes:
        attrs_name = attrs.argname.upper()
        field_to_human = attrs_field_to_human.get(attrs.name, dict())
        for attr_field in attrs._fields:
            if attr_field not in attrs.internal:
                attr_flag = field_to_human.get(attr_field, False)
                const_to_human_attrs_flag['_'.join((FTL_ATTR, attrs_name, attr_field.upper(), 'HUMAN'))] = attr_flag

    # Prepare module
    module_path = module.split('.', 1)[0]
    modules_patched = set()

    # Prepare patch hash value
    patch_hash = hash(tuple(hash(tuple(sorted(c.items())))
                            for c in (const_to_record_flag, const_to_attrs_idx, const_to_human_attrs_flag)))

    # Prepare module patcher
    def patch_module(module_name):
        """ Recursively patch spec-related constants in parser (sub-)modules.
        """
        # Prevent patching the same module multiple times
        modules_patched.add(module_name)

        # Access module instance
        module = sys.modules[module_name]

        # Patch module members
        for member in sorted(dir(module)):

            # Patch record flags
            if member in const_to_record_flag:
                setattr(module, member, const_to_record_flag[member])

            # Patch attribute indexes
            elif member in const_to_attrs_idx:
                setattr(module, member, const_to_attrs_idx[member])

            # Patch human-readable attribute flags
            elif member in const_to_human_attrs_flag:
                setattr(module, member, const_to_human_attrs_flag[member])

            # Recurse down
            sub_module = getattr(getattr(module, member), '__module__', None)
            if sub_module is not None and sub_module.startswith(f'{module_path}.'):
                if sub_module not in modules_patched:
                    patch_module(sub_module)
                    continue

    # Lock module access based on patch hash value
    with FTL_PATCH_THREAD_LOCK:
        if getattr(sys.modules[module], FTL_PATCH_HASH, patch_hash) != patch_hash:
            raise FtlError(f'Unable to lock parser module "{module}" - parallel use not allowed')
        setattr(sys.modules[module], FTL_PATCH_LOCKED, getattr(sys.modules[module], FTL_PATCH_LOCKED, 0) + 1)
        setattr(sys.modules[module], FTL_PATCH_HASH, patch_hash)

        # Patch all modules
        patch_module(module)


def unpatch_spec_records(module):
    """ Undo module patch (remove lock).
    """
    # Unlock module access
    with FTL_PATCH_THREAD_LOCK:
        setattr(sys.modules[module], FTL_PATCH_LOCKED, getattr(sys.modules[module], FTL_PATCH_LOCKED, 0) - 1)
        if getattr(sys.modules[module], FTL_PATCH_LOCKED, 0) == 0:
            delattr(sys.modules[module], FTL_PATCH_LOCKED)
            delattr(sys.modules[module], FTL_PATCH_HASH)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
# flake8: noqa
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from .util import generate_spec_record

# Available BGP records
FTL_RECORD_BGP_PEER_TABLE    = 1 << 0
FTL_RECORD_BGP_STATE_CHANGE  = 1 << 1
FTL_RECORD_BGP_ROUTE         = 1 << 2
FTL_RECORD_BGP_KEEP_ALIVE    = 1 << 3
FTL_RECORD_BGP_ROUTE_REFRESH = 1 << 4
FTL_RECORD_BGP_NOTIFICATION  = 1 << 5
FTL_RECORD_BGP_OPEN          = 1 << 6
FTL_RECORD_BGP_STATS         = 1 << 7
FTL_RECORD_BGP_ERROR         = 1 << 8

# Availabe BGP record names
FTL_RECORD_BGP_PEER_TABLE_NAME    = 'peer_table'
FTL_RECORD_BGP_STATE_CHANGE_NAME  = 'state_change'
FTL_RECORD_BGP_ROUTE_NAME         = 'route'
FTL_RECORD_BGP_KEEP_ALIVE_NAME    = 'keep_alive'
FTL_RECORD_BGP_ROUTE_REFRESH_NAME = 'route_refresh'
FTL_RECORD_BGP_NOTIFICATION_NAME  = 'notification'
FTL_RECORD_BGP_OPEN_NAME          = 'open'
FTL_RECORD_BGP_STATS_NAME         = 'stats'
FTL_RECORD_BGP_ERROR_NAME         = 'error'

# BGP record specification
# [Format] spec := (field, value, default)
FtlRecordsBgp = generate_spec_record('FtlRecordsBgp', (
    (FTL_RECORD_BGP_PEER_TABLE_NAME,    FTL_RECORD_BGP_PEER_TABLE,    False),
    (FTL_RECORD_BGP_STATE_CHANGE_NAME,  FTL_RECORD_BGP_STATE_CHANGE,  False),
    (FTL_RECORD_BGP_ROUTE_NAME,         FTL_RECORD_BGP_ROUTE,         True),
    (FTL_RECORD_BGP_KEEP_ALIVE_NAME,    FTL_RECORD_BGP_KEEP_ALIVE,    False),
    (FTL_RECORD_BGP_ROUTE_REFRESH_NAME, FTL_RECORD_BGP_ROUTE_REFRESH, False),
    (FTL_RECORD_BGP_NOTIFICATION_NAME,  FTL_RECORD_BGP_NOTIFICATION,  False),
    (FTL_RECORD_BGP_OPEN_NAME,          FTL_RECORD_BGP_OPEN,          False),
    (FTL_RECORD_BGP_STATS_NAME,         FTL_RECORD_BGP_STATS,         True),
    (FTL_RECORD_BGP_ERROR_NAME,         FTL_RECORD_BGP_ERROR,         True),
), human=False)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
# flake8: noqa
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import socket
import struct
from datetime import datetime

# IP protocols
IPV4 = 4
IPV6 = 6

# IP protocol strings
IPV4_STR = 'ipv4'
IPV6_STR = 'ipv6'

# Address families
AF_INET = int(socket.AF_INET)
AF_INET6 = int(socket.AF_INET6)

# AFI values
AFI_IPV4 = 1  # Defined by IANA
AFI_IPV6 = 2  # Defined by IANA

# Socket/struct functions
socket_inet_pton = socket.inet_pton
socket_inet_ntop = socket.inet_ntop
struct_unpack = struct.unpack
struct_pack = struct.pack

# Struct bytes
STRUCT_2B = '!H'
STRUCT_4B = '!I'
STRUCT_8B = '!Q'
STRUCT_2B2B = '!HH'
STRUCT_8B8B = '!QQ'
STRUCT_2B2B2B = '!HHH'
STRUCT_4B4B4B = '!III'

# Strings
UTF8 = 'utf-8'

# Datetime functions
datetime_utcfromtimestamp = datetime.utcfromtimestamp

# Datetime formats
DATETIME_FORMAT_USEC = '%Y-%m-%d %H:%M:%S.%f'
DATETIME_FORMAT_MIN = '%Y-%m-%d %H:%M'



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import bz2
import contextlib
import functools
import gzip
import time
from collections import namedtuple

# Local imports
from .model.record import FtlRecordsBgp
from .model.error import FtlError
from .model.error import FtlFileError
from .model.error import FtlFormatError
from .model.const import FTL_MRT
from .model.const import FTL_LGL
from .model.const import FTL_BGP
from .model.const import FTL_PARSER
from .model.util import apply_spec_records
from .model.util import patch_spec_records
from .model.util import unpatch_spec_records
from .model.stats import filter_stats_spec
from .model.stats import init_stats_record
from .data.util import handle_bgp_error
from .data.util import init_caches

# Magic bytes
BZ2_MAGIC_BYTES = b'\x42\x5a\x68'
GZIP_MAGIC_BYTES = b'\x1f\x8b'


def FtlParser(*unpack_functions):
    """ BGP parser decorator for applying predefined arguments to given unpack functions.

        Positional arguments:
          unpack_functions  - One or more parsing functions decorated with @FtlParserFunc.
    """
    # Check setup of unpack functions
    for unpack_function in unpack_functions:
        if hasattr(unpack_function, 'ftl_parser_open_file') is False:
            raise FtlError(f'Parser function {unpack_function.__name__}() invalid (needs @FtlParserFunc decorator)')

    # Prepare supported data sources
    parser_support = {FTL_PARSER, FTL_BGP}
    parser_support |= {f.__name__.split('_')[1] for f in unpack_functions} & {FTL_MRT, FTL_LGL}
    parser_no_support = {FTL_MRT, FTL_LGL} - parser_support

    # Prepare supported stats fields
    FtlAttrsBgpSupport = filter_stats_spec(parser_no_support)

    # Initialize parser context
    def decorator(parser):
        """ Setup generic BGP parser.
        """
        @functools.wraps(parser, assigned=set(functools.WRAPPER_ASSIGNMENTS) - {'__doc__'})
        def ftl_parser(named_records=True, serialize=False, use_cache=True, raise_on_errors=False,
                       bgp_records=FtlRecordsBgp.DEFAULT, **bgp_attributes):
            """ This @FtlParser instance is used to read input files and generate a set of BGP records (Python tuples).
                It must be used with a context manager (see sample usage below) and accepts the following arguments.

                Keyword arguments:
                  named_records     - Return named tuples instead of plain unnamed tuple records. [Default: True]
                  serialize         - Convert output records to JSON (if named) or CSV (if unnamed). [Default: False]
                  use_cache         - Cache expensive low-entropy values (memory consumption < 1MB). [Default: True]
                  raise_on_errors   - Raise exceptions and stop parsing in case of data errors. [Default: False]
                  bgp_records       - Select the types of BgpRecord entries to be returned by the parser.
                                      Supports bitwise logical operators OR, AND, NOT to specify multiple records.
                                      {BGP_RECORDS_DOC}
                  {BGP_ATTRIBUTES_DOC}

                Returns:
                  parse(<filename>) - Parse function that accepts a filename as single positional argument and generates
                                      all specified BGP records on invocation. Input files can be provided in .bz2 or .gz
                                      format. The parse function may be invoked multiple times within a single context.

                Raises:
                  FtlError          - Generic parser or runtime error.
                  FtlFileError      - Failure during access of input file.
                  FtlFormatError    - Unexpected format of input file.
                  FtlDataError      - Invalid data entry in input file.

                Sample usage:

                  # Import parser
                  from ftlbgp import BgpParser

                  # Parse all records
                  filename = ...
                  with BgpParser(bgp_records=BgpParser.bgp.records.ALL) as parse:
                      for record in parse(filename):
                          ...

                  # Parse specific records (route and error)
                  filename = ...
                  with BgpParser(bgp_records=BgpParser.bgp.records.route | BgpParser.bgp.records.error) as parse:
                      for record in parse(filename):
                          ...

                  # Parse all route attributes
                  filename = ...
                  with BgpParser(bgp_route=BgpParser.bgp.route.ALL) as parse:
                      for record in parse(filename):
                          ...

                  # Parse specific route attributes (default and local_pref)
                  filename = ...
                  with BgpParser(bgp_route=BgpParser.bgp.route.DEFAULT | BgpParser.bgp.route.local_pref) as parse:
                      for record in parse(filename):
                          ...
            """
            # Prepare total runtime of context maanager
            parser_time_start = time.perf_counter()

            # Prepare record attributes and add default values
            bgp_attrs = {attrs.name: bgp_attributes.pop(attrs.argname, attrs.DEFAULT)
                         for attrs in FtlAttrsBgpSupport if bgp_records & getattr(FtlRecordsBgp, attrs.name)}

            # Set empty record attributes for other records
            bgp_attrs.update({attrs.name: attrs.NONE for attrs in FtlAttrsBgpSupport
                              if not bgp_records & getattr(FtlRecordsBgp, attrs.name)})

            # Check remaining keywords
            if len(bgp_attributes) > 0:
                bgp_attribute = next(iter(bgp_attributes))
                if any(attrs.argname == bgp_attribute for attrs in FtlAttrsBgpSupport) is True:
                    raise FtlError(f'Argument "{bgp_attribute}" not allowed for parser - enable via "bgp_records"'
                                   '(see @FtlParser decorator)')
                raise FtlError(f'Invalid argument "{bgp_attribute}" for parser (see @FtlParser decorator)')

            # Create record attributes instance
            bgp_attributes = namedtuple('FtlRecordAttrsBgp', bgp_attrs.keys())(**bgp_attrs)

            # Prepare BGP record templates
            (bgp_templates, bgp_error, records_fields, attrs_fields, attrs_human, stats_emit, stats_record,
             ) = apply_spec_records('Bgp', named_records, serialize, raise_on_errors, handle_bgp_error, FtlRecordsBgp,
                                    FtlAttrsBgpSupport, bgp_records, bgp_attributes)

            # Patch BGP record specs
            # NOTE: This call patches all spec-related constants that are imported
            # NOTE: by the unpack functions' modules (including any child function modules)
            for unpack_function in unpack_functions:
                patch_spec_records(unpack_function.__module__, FtlRecordsBgp, FtlAttrsBgpSupport, records_fields,
                                   attrs_fields, attrs_human)

            # Prepare record values and attr indexes
            records_name, records_cls = FtlRecordsBgp.spec, FtlRecordsBgp.__class__.__name__
            records_spec = namedtuple(records_cls, records_fields.keys())(*records_fields.values())
            attrs_specs = dict()
            for attrs_name, attrs in attrs_fields.items():
                attrs_cls = getattr(FtlAttrsBgpSupport, attrs_name).__class__.__name__
                attrs_specs[attrs_name] = namedtuple(attrs_cls, attrs.keys())(*attrs.values())
            unpack_bgp_spec = namedtuple('FtlBgp', (records_name, ) + tuple(attrs_specs))
            unpack_bgp_spec = unpack_bgp_spec(records_spec, *attrs_specs.values())

            # Prepare human-readable caches
            # NOTE: At the moment, there is only one simple timestamp cache (CACHE_TS)
            # NOTE: We could easily introduce additional cache types (e.g. for CACHE_IP)
            # NOTE: In that case, we should also add some cache fill stats below
            unpack_caches = [init_caches() if use_cache else None for _ in range(len(unpack_functions))]

            # Initialize stats record
            if bgp_records & FtlRecordsBgp.stats:
                init_stats_record(FtlAttrsBgpSupport.stats, unpack_bgp_spec.stats, bgp_attributes.stats, stats_record)

            @contextlib.contextmanager
            def parse():
                """ Context manager for BGP parsing.
                """
                def unpack(filename):
                    """ Try to parse input file with given unpack functions.
                    """
                    # Iterate unpack functions
                    idx, max_idx = 0, len(unpack_functions) - 1
                    for idx, (unpack_function, caches) in enumerate(zip(unpack_functions, unpack_caches)):
                        source_runtime = f'{unpack_function.__name__.split("_")[1]}_runtime'
                        unpack_time_start = time.perf_counter()

                        try:
                            # Invoke file opener
                            with unpack_function.ftl_parser_open_file(filename) as inputfile:

                                # Yield unpacked records
                                yield from unpack_function(inputfile, caches, stats_record, bgp_templates, bgp_error)

                            # Success
                            return

                        # Handle format errors
                        except FtlFormatError:

                            # Retry if unpack functions left
                            if idx < max_idx:
                                continue

                            # Raise generic error if multiple parsers failed
                            if max_idx > 0:
                                raise FtlFormatError('Unknown input file format')  # pylint: disable=raise-missing-from

                            # Reraise specific format error
                            raise

                        # Reraise already handled exceptions
                        except FtlError:
                            raise

                        # Raise unhandled exceptions
                        except Exception as exc:
                            raise FtlError('Unhandled parser error', exception=exc)  # pylint: disable=raise-missing-from

                        # Update and yield stats record
                        finally:

                            # Update stats record
                            if bgp_records & FtlRecordsBgp.stats:
                                now_time_end = time.perf_counter()
                                unpack_time = now_time_end - unpack_time_start
                                parser_time = now_time_end - parser_time_start
                                sattrs = FtlAttrsBgpSupport.stats

                                # Add runtimes
                                if (bgp_attributes.stats
                                    & (getattr(sattrs, source_runtime) | getattr(sattrs.human, source_runtime))):
                                    stats_record[getattr(unpack_bgp_spec.stats, source_runtime)] += unpack_time
                                if bgp_attributes.stats & (sattrs.parser_runtime | sattrs.human.parser_runtime):
                                    stats_record[unpack_bgp_spec.stats.parser_runtime] += unpack_time
                                if bgp_attributes.stats & (sattrs.parser_lifetime | sattrs.human.parser_lifetime):
                                    stats_record[unpack_bgp_spec.stats.parser_lifetime] = parser_time

                                # Yield final stats record
                                # NOTE: This record is cumulatively updated when parsing multiple files
                                yield stats_emit(stats_record)

                # Add record values and attr indexes
                unpack.bgp = unpack_bgp_spec

                # Yield record generator
                yield unpack

                # Unpatch BGP record specs
                for unpack_function in unpack_functions:
                    unpatch_spec_records(unpack_function.__module__)

            # Return parser context
            return parse()

        # Prepare parser documentation
        line_indent = ''.join(('\n', ftl_parser.__doc__.split('\n')[-1], ' ' * 6))
        arg_indent = ' ' * (len(next(line for line in ftl_parser.__doc__.split('\n')
                                     if FtlRecordsBgp.argname in line).split('-', 1)[0].lstrip()))

        def generate_doc(model, records=False):
            """ Generate docstring for given record/attributes model.
            """
            # Extract model defaults and available entries
            model_doc_default, model_doc_available, model_doc_internal = list(), list(), list()
            for field, value in model._asdict().items():
                if field not in model.internal or any(v is value for v in (model.ALL, model.NONE, model.DEFAULT)):
                    model_doc = model_doc_available
                    if value in {model.ALL, model.NONE, model.DEFAULT}:
                        model_doc = model_doc_internal
                    elif value & model.DEFAULT:
                        model_doc = model_doc_default
                    field_name = f'{parser.__name__}{".bgp" if records is False else ""}.{model.name}'
                    field_name += f'{".records" if records is True else "[.human]"}.{field}'
                    field_doc = ''.join((line_indent, arg_indent, ' ' * 4, field_name))
                    model_doc.append(field_doc)

            # Generate and return model documentation
            return ''.join(['[Default:', ' |'.join(model_doc_default), line_indent, arg_indent, ' ' * 2, ' Available:',
                            ''.join(model_doc_available + model_doc_internal).rstrip('|'), ']'])

        # Prepare records documentation
        records_doc = generate_doc(FtlRecordsBgp, records=True).rstrip()

        # Prepare attributes documentation
        attributes_doc = [line_indent[1:]]
        for name, attrs in FtlAttrsBgpSupport._asdict().items():
            record_name = f'Bgp{"".join((p.capitalize() for p in name.split("_")))}Record'
            attributes_doc.append(''.join((
                line_indent, attrs.argname, ' ' * (len(arg_indent) - len(attrs.argname)),
                f'- Select [optionally human-readable] attribute to be included in {record_name} entries.',
                line_indent, ' ' * (len(attrs.argname) + 2 + len(arg_indent) - len(attrs.argname)),
                'Supports bitwise logical operators (OR/AND/NOT) to specify multiple attributes.', line_indent,
                ' ' * len(attrs.argname), ' ' * (len(arg_indent) - len(attrs.argname)), ' ' * 2, generate_doc(attrs)
            )))

        # Update parser documentation
        attributes_doc_str = ''.join(attributes_doc).lstrip()
        ftl_parser.__doc__ = ftl_parser.__doc__.format(BGP_RECORDS_DOC=records_doc, BGP_ATTRIBUTES_DOC=attributes_doc_str)
        ftl_parser.__doc__ = ''.join((parser.__doc__.strip(), line_indent, line_indent[:-2], ftl_parser.__doc__.strip()))

        # Prepare spec filtering (generate key/value spec lists without internal fields)
        # pylint: disable-next=unnecessary-lambda-assignment
        filter_spec_record = lambda s: list(zip(*((k, v) for k, v in s._asdict().items() if k not in s.internal
                                                  or any(vs is v for vs in (s.ALL, s.NONE, s.DEFAULT)))))
        # pylint: disable-next=unnecessary-lambda-assignment
        filter_spec_attrs = lambda s: list(zip(*((k, v) for k, v in s._asdict().items() if k not in s.internal
                                                 or any(vs is v for vs in (s.ALL, s.NONE, s.DEFAULT, s.human)))))

        # Add record/attr specs (without internal fields)
        records_name, (records_fields, records_values) = FtlRecordsBgp.spec, filter_spec_record(FtlRecordsBgp)
        records_spec = namedtuple(FtlRecordsBgp.__class__.__name__, records_fields)(*records_values)
        attrs_specs = dict()
        for attrs_name, attrs in FtlAttrsBgpSupport._asdict().items():
            attrs_fields, attrs_values = filter_spec_attrs(attrs)
            attrs_specs[attrs_name] = namedtuple(attrs.__class__.__name__, attrs_fields)(*attrs_values)
        ftl_parser.bgp = namedtuple('FtlBgp', (records_name, ) + tuple(attrs_specs))(records_spec, *attrs_specs.values())

        # Return parser
        return ftl_parser

    # Return decorator for parser
    return decorator


def FtlParserFunc(method=None, text_input=None):
    """ Decorate BGP parsing function.

        Keyword arguments:
          text_input  - If given, open input file as text with specified encoding (e.g. UTF-8).
    """
    def decorator(unpack):
        """ Setup generic BGP parsing function.
        """
        @functools.wraps(unpack)
        def ftl_parser_open_file(filename):
            """ Open BGP file for reading.
            """
            try:
                # Read magic bytes
                ftype = ''
                with open(filename, 'rb') as fh:
                    ftype = fh.read(max(len(BZ2_MAGIC_BYTES), len(GZIP_MAGIC_BYTES)))

                # Prepare access mode
                mode = 'rb' if text_input is None else 'rt'

                # Support bzip2, gzip, and uncompressed files
                if ftype.startswith(BZ2_MAGIC_BYTES):
                    return bz2.open(filename, mode=mode, encoding=text_input)
                if ftype.startswith(GZIP_MAGIC_BYTES):
                    return gzip.open(filename, mode=mode, encoding=text_input)
                return open(filename, mode=mode, encoding=text_input)

            # Invalid input file
            except Exception as exc:
                raise FtlFileError(exception=exc)  # pylint: disable=raise-missing-from

            # Failure
            return None

        # Add file opener to parser function
        unpack.ftl_parser_open_file = ftl_parser_open_file

        # Return parser function
        return unpack

    # Return decorator for parser function
    if method is None:
        return decorator
    return decorator(method)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# Local imports
from ..bgp.attr import unpack_mrt_bgp_attr
from ...util import CACHE_TS
from ...const import IPV4
from ...const import IPV6
from ...const import IPV4_STR
from ...const import IPV6_STR
from ...const import AF_INET
from ...const import AF_INET6
from ...const import AFI_IPV4
from ...const import AFI_IPV6
from ...const import STRUCT_2B
from ...const import STRUCT_4B
from ...const import STRUCT_8B8B
from ...const import DATETIME_FORMAT_USEC
from ...const import DATETIME_FORMAT_MIN
from ...const import struct_unpack
from ...const import socket_inet_ntop
from ...const import datetime_utcfromtimestamp
from ....model.attr import FTL_ATTR_BGP_ROUTE_SOURCE
from ....model.attr import FTL_ATTR_BGP_ROUTE_SEQUENCE
from ....model.attr import FTL_ATTR_BGP_ROUTE_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_AS
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_IP
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX
from ....model.attr import FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4
from ....model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6
from ....model.const import FTL_ATTR_BGP_ROUTE_SOURCE_RIB
from ....model.const import FTL_ATTR_BGP_ROUTE_SOURCE_RIB_STR
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL
from ....model.record import FTL_RECORD_BGP_ROUTE
from ....model.record import FTL_RECORD_BGP_STATS
from ....model.error import FtlMrtDataError


def unpack_mrt_entry_td_rib(caches, stats_record, route_records, entry_bytes, mrt_subtype):
    """ Parse MRT table dump entry.
    """
    # ------------------------------
    # [RFC6396] 4.2. TABLE_DUMP Type
    # ------------------------------
    # The Subtype field is used to encode whether the RIB entry contains
    # IPv4 or IPv6 [RFC2460] addresses. There are two possible values for
    # the Subtype as shown below.
    #
    #     1    AFI_IPv4
    #     2    AFI_IPv6
    #
    #  0                   1                   2                   3
    #  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |         View Number           |       Sequence Number         |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                        Prefix (variable)                      |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # | Prefix Length |    Status     |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                         Originated Time                       |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                    Peer IP Address (variable)                 |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |           Peer AS             |       Attribute Length        |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                   BGP Attribute... (variable)
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Check AFI value
    if (mrt_subtype != AFI_IPV4 and mrt_subtype != AFI_IPV6):  # pylint: disable=consider-using-in
        raise FtlMrtDataError(f'Unknown AFI value ({mrt_subtype}) in table dump entry',
                              reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=entry_bytes)

    # Access route record template
    route_init, route_emit, route_error = route_records

    # Initialize route record
    route_record = list(route_init)

    # Add source to route record
    if FTL_ATTR_BGP_ROUTE_SOURCE >= 0:
        if FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN:
            route_record[FTL_ATTR_BGP_ROUTE_SOURCE] = FTL_ATTR_BGP_ROUTE_SOURCE_RIB_STR
        else:
            route_record[FTL_ATTR_BGP_ROUTE_SOURCE] = FTL_ATTR_BGP_ROUTE_SOURCE_RIB

    # Add prefix protocol to route record
    afinet = AF_INET6 if mrt_subtype == AFI_IPV6 else AF_INET
    if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL >= 0:
        if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL_HUMAN:
            route_record[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = IPV6_STR if afinet == AF_INET6 else IPV4_STR
        else:
            route_record[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = IPV6 if afinet == AF_INET6 else IPV4

    # Prepare byte offset
    offset = 0

    # Skip view number
    offset += 2

    # Add sequence number to route record
    if FTL_ATTR_BGP_ROUTE_SEQUENCE >= 0:
        route_record[FTL_ATTR_BGP_ROUTE_SEQUENCE] = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
    offset += 2

    # Add prefix protocol to route record
    if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL >= 0:
        if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL_HUMAN:
            route_record[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = IPV6_STR if afinet == AF_INET6 else IPV4_STR
        else:
            route_record[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = IPV6 if afinet == AF_INET6 else IPV4

    # Parse prefix bytes
    preflen = 16 if afinet == AF_INET6 else 4
    prefix = entry_bytes[offset:offset + preflen]
    if FTL_ATTR_BGP_ROUTE_PREFIX_HUMAN:
        prefix = socket_inet_ntop(afinet, prefix)
    elif afinet == AF_INET6:
        net, host = struct_unpack(STRUCT_8B8B, prefix)
        prefix = (net << 64) + host
    else:
        prefix = struct_unpack(STRUCT_4B, prefix)[0]
    offset += preflen

    # Parse prefix mask
    mask = entry_bytes[offset]
    offset += 1

    # Add prefix to route record
    if FTL_ATTR_BGP_ROUTE_PREFIX >= 0:
        if FTL_ATTR_BGP_ROUTE_PREFIX_HUMAN:
            route_record[FTL_ATTR_BGP_ROUTE_PREFIX] = f'{prefix}/{mask}'
        else:
            route_record[FTL_ATTR_BGP_ROUTE_PREFIX] = (prefix, mask)

    # Skip status
    offset += 1

    # Parse timestamp
    if FTL_ATTR_BGP_ROUTE_TIMESTAMP >= 0:
        ts = float(struct_unpack(STRUCT_4B, entry_bytes[offset:offset + 4])[0])
        if FTL_ATTR_BGP_ROUTE_TIMESTAMP_HUMAN:

            # Cache date string for minute-based timestamp
            if caches:
                ts_cache = caches[CACHE_TS]
                ts_min, ts_sec = divmod(ts, 60)
                ts_cached = ts_cache.get(ts_min, None)
                if ts_cached is None:
                    ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                    ts_cache[ts_min] = ts_cached

                # Add seconds and microseconds
                ts = f'{ts_cached}:{ts_sec:09.6f}'

            # Do not use cache
            else:
                ts = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_USEC)

        # Add timestamp to route record
        route_record[FTL_ATTR_BGP_ROUTE_TIMESTAMP] = ts
    offset += 4

    # Add peer protocol to route record
    if FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL >= 0:
        if FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN:
            route_record[FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL] = (IPV6_STR if afinet == AF_INET6 else IPV4_STR)
        else:
            route_record[FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL] = IPV6 if afinet == AF_INET6 else IPV4

    # Add peer IP to route record
    iplen = 16 if afinet == AF_INET6 else 4
    peer_ip = entry_bytes[offset:offset + iplen]
    if FTL_ATTR_BGP_ROUTE_PEER_IP_HUMAN:
        peer_ip = socket_inet_ntop(afinet, peer_ip)
    elif afinet == AF_INET6:
        net, host = struct_unpack(STRUCT_8B8B, peer_ip)
        peer_ip = (net << 64) + host
    else:
        peer_ip = struct_unpack(STRUCT_4B, peer_ip)[0]
    if FTL_ATTR_BGP_ROUTE_PEER_IP >= 0:
        route_record[FTL_ATTR_BGP_ROUTE_PEER_IP] = peer_ip
    offset += iplen

    # Add peer AS to route record
    peer_as = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
    if FTL_ATTR_BGP_ROUTE_PEER_AS >= 0:
        route_record[FTL_ATTR_BGP_ROUTE_PEER_AS] = peer_as
    offset += 2

    # Parse attributes length
    alen = struct_unpack(STRUCT_2B, entry_bytes[offset:offset + 2])[0]
    offset += 2

    # Parse attributes
    attr_bytes = entry_bytes[offset:offset + alen]
    offset += alen
    try:
        unpack_mrt_bgp_attr(caches, stats_record, route_init, route_emit, route_record, attr_bytes, aslen=2, rib=True)

    # Yield (or re-raise) attribute errors
    except FtlMrtDataError as error:
        yield from route_error(error)
        return

    # Update stats record
    if FTL_RECORD_BGP_STATS:

        # Update IPv4 RIB routes
        if afinet == AF_INET:
            if FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4 >= 0:
                stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4] += 1

        # Update IPv6 RIB routes
        if afinet == AF_INET6:
            if FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6 >= 0:
                stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6] += 1

    # Yield final route record
    if FTL_RECORD_BGP_ROUTE:
        yield route_emit(route_record)



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import re

# Local imports
from .table import check_lgl_table
from ..const import IPV4
from ..const import IPV6
from ..const import IPV4_STR
from ..const import IPV6_STR
from ..const import AF_INET
from ..const import AF_INET6
from ..const import STRUCT_4B
from ..const import STRUCT_8B8B
from ..const import socket_inet_pton
from ..const import struct_unpack
from ...parser import FtlParserFunc
from ...model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL
from ...model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID
from ...model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_AS
from ...model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL_HUMAN
from ...model.attr import FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID_HUMAN
from ...model.attr import FTL_ATTR_BGP_ROUTE_SOURCE
from ...model.attr import FTL_ATTR_BGP_ROUTE_SEQUENCE
from ...model.attr import FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL
from ...model.attr import FTL_ATTR_BGP_ROUTE_PEER_BGP_ID
from ...model.attr import FTL_ATTR_BGP_ROUTE_PEER_AS
from ...model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL
from ...model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_IP
from ...model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL
from ...model.attr import FTL_ATTR_BGP_ROUTE_PREFIX
from ...model.attr import FTL_ATTR_BGP_ROUTE_ASPATH
from ...model.attr import FTL_ATTR_BGP_ROUTE_ORIGIN
from ...model.attr import FTL_ATTR_BGP_ROUTE_LOCAL_PREF
from ...model.attr import FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC
from ...model.attr import FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN
from ...model.attr import FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN
from ...model.attr import FTL_ATTR_BGP_ROUTE_PEER_BGP_ID_HUMAN
from ...model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL_HUMAN
from ...model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_IP_HUMAN
from ...model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL_HUMAN
from ...model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_HUMAN
from ...model.attr import FTL_ATTR_BGP_ROUTE_ORIGIN_HUMAN
from ...model.attr import FTL_ATTR_BGP_STATS_LGL_ENTRIES
from ...model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4
from ...model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6
from ...model.const import FTL_ATTR_BGP_ROUTE_SOURCE_RIB
from ...model.const import FTL_ATTR_BGP_ROUTE_SOURCE_RIB_STR
from ...model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_IGP
from ...model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_IGP_STR
from ...model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_EGP
from ...model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_EGP_STR
from ...model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE
from ...model.const import FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE_STR
from ...model.const import FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA
from ...model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR
from ...model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_ASPATH
from ...model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX
from ...model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_AS
from ...model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_IP
from ...model.record import FTL_RECORD_BGP_PEER_TABLE
from ...model.record import FTL_RECORD_BGP_ROUTE
from ...model.record import FTL_RECORD_BGP_STATS
from ...model.error import FtlLglError
from ...model.error import FtlLglHeaderError
from ...model.error import FtlLglFormatError
from ...model.error import FtlLglDataError

# Header parts
HEADER_PEER_AS = 'local AS'
HEADER_PEER_BGP_ID = 'local router ID is'
HEADER_LOCAL_PREF = 'Default local pref'

# Route totals regex
ROUTE_TOTAL1_REGEX = re.compile(r'^Displayed\s+(\d+) routes and\s+(\d+) total paths$')
ROUTE_TOTAL2_REGEX = re.compile(r'^Total number of prefixes (\d+)$')
ROUTE_HEADER_REGEX = re.compile(r'^View \S+ \S+ (\d+) routes$')

# Multiline routes
ROUTE_MAX_MULTILINE_LEN = 3


@FtlParserFunc(text_input='utf-8')
# pylint: disable-next=unused-argument
def unpack_lgl_data(inputfile, caches, stats_record, bgp_records, bgp_error):
    """ Parse looking glass RIB dumps in plain text format.
    """
    # Access BGP record templates
    route_init, route_emit, route_error = bgp_records.route
    peer_table_init, peer_table_emit, peer_table_error = bgp_records.peer_table

    # Initialize route record header
    route_record_header = list(route_init)

    # Add source to route record header
    if FTL_ATTR_BGP_ROUTE_SOURCE >= 0:
        source = FTL_ATTR_BGP_ROUTE_SOURCE_RIB_STR if FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN else FTL_ATTR_BGP_ROUTE_SOURCE_RIB
        route_record_header[FTL_ATTR_BGP_ROUTE_SOURCE] = source

    ##############
    # BGP HEADER #
    ##############

    # ------------------------
    # [PCH] % sh bgp ipv4 wide
    # ------------------------
    # BGP table version is 147996423, local router ID is 74.80.112.4, vrf id 0
    # Default local pref 100, local AS 3856
    # Status codes:  s suppressed, d damped, h history, u unsorted, * valid, > best, = multipath,
    #                i internal, r RIB-failure, S Stale, R Removed
    # Nexthop codes: @NNN nexthop's vrf id, < announce-nh-self
    # Origin codes:  i - IGP, e - EGP, ? - incomplete
    #
    #    Network          Next Hop            Metric LocPrf Weight Path

    # -------------------------
    # [RIPE] % sh bgp ipv4 wide
    # -------------------------
    # View #0 inet 64059 routes
    # Status code: s suppressed, * valid, > best, i - internal
    # Origin codes: i - IGP, e - EGP, ? - incomplete, a - aggregate
    #
    #   P Pref Time     Destination                Next Hop                 If      Path

    # Prepare dump data
    line = None

    try:
        # Prepare dump status
        header_errors = list()

        # Prepare details
        peer_as, peer_bgp_id = None, None
        prefixes, routes = None, None

        # Parse dump header
        try:
            for line in inputfile:

                # Extract totals from first line for sanitarization
                match = ROUTE_HEADER_REGEX.match(line)
                if match is not None:
                    try:
                        prefixes = int(match.group(1))
                    except ValueError:
                        pass

                # Extract default local-pref
                if HEADER_LOCAL_PREF in line:
                    try:
                        # Try to parse local-pref
                        locpref = int(line.split(HEADER_LOCAL_PREF, 1)[1].strip().split(None, 1)[0].rstrip(',').strip())

                        # Add local-pref to route record header
                        if FTL_ATTR_BGP_ROUTE_LOCAL_PREF >= 0:
                            route_record_header[FTL_ATTR_BGP_ROUTE_LOCAL_PREF] = locpref

                    # Store header error
                    except Exception as exc:  # pylint: disable=broad-except
                        header_errors.append((route_error, FtlLglDataError('Unable to decode local-pref',
                                              reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR, data=line, exception=exc)))

                # Extract peer AS
                if HEADER_PEER_AS in line:
                    try:
                        # Try to parse peer AS
                        peer_as = int(line.split(HEADER_PEER_AS, 1)[1].strip().split(None, 1)[0].rstrip(',').strip())

                        # Add peer AS to route record header
                        if FTL_ATTR_BGP_ROUTE_PEER_AS >= 0:
                            route_record_header[FTL_ATTR_BGP_ROUTE_PEER_AS] = peer_as

                    # Store header error
                    except Exception as exc:  # pylint: disable=broad-except
                        header_errors.append((peer_table_error, FtlLglDataError('Unable to decode peer AS',
                                              reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_AS, data=line, exception=exc)))

                # Extract peer BGP ID (IPv4 only)
                if HEADER_PEER_BGP_ID in line:
                    try:
                        # Try to parse peer BGP ID
                        peer_bgp_id = line.split(HEADER_PEER_BGP_ID, 1)[1].strip().split(None, 1)[0].rstrip(',').strip()
                        peer_bgp_id_int = struct_unpack(STRUCT_4B, socket_inet_pton(AF_INET, peer_bgp_id))[0]

                        # Add peer BGP ID to route record header
                        peer_bgp_id = peer_bgp_id if FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID_HUMAN else peer_bgp_id_int
                        if FTL_ATTR_BGP_ROUTE_PEER_BGP_ID >= 0:
                            peer_bgp_id_route = peer_bgp_id if FTL_ATTR_BGP_ROUTE_PEER_BGP_ID_HUMAN else peer_bgp_id_int
                            route_record_header[FTL_ATTR_BGP_ROUTE_PEER_BGP_ID] = peer_bgp_id_route

                    # Store header error
                    except Exception as exc:  # pylint: disable=broad-except
                        header_errors.append((peer_table_error, FtlLglDataError('Unable to decode peer BGP ID',
                                              reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_IP, data=line, exception=exc)))

                # Check for valid table start
                unpack_lgl_entry = check_lgl_table(line)
                if unpack_lgl_entry is not None:
                    break

        # Always raise critical file reading errors
        except Exception as exc:
            # pylint: disable-next=raise-missing-from
            raise FtlLglHeaderError('Unknown lgl format', data=line, exception=exc)

        # Always raise critical header errors
        if unpack_lgl_entry is None:
            raise FtlLglHeaderError('No BGP table found', data=line)

        # Yield (or re-raise) remaining header errors
        for record_error, error in header_errors:
            yield from record_error(error)

        # Prepare peer table record
        peer_table_record = None
        if peer_as is not None or peer_bgp_id is not None:

            # Initialize peer table record
            peer_table_record = list(peer_table_init)

            # Add peer AS to peer table record
            if peer_as is not None:
                if FTL_ATTR_BGP_PEER_TABLE_PEER_AS >= 0:
                    peer_table_record[FTL_ATTR_BGP_PEER_TABLE_PEER_AS] = peer_as

            # Add peer BGP ID to peer table record
            if peer_bgp_id is not None:
                if FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID >= 0:
                    peer_table_record[FTL_ATTR_BGP_PEER_TABLE_PEER_BGP_ID] = peer_bgp_id

        #############
        # BGP TABLE #
        #############

        # Prepare record parser
        n_routes4, n_routes6, n_prefixes = 0, 0, 0

        # Initialize route record and parser state
        route_record_prefix, route_record_proto, last_prefix = None, None, None
        multiline, multiline_len = '', 0

        # Iterate input lines
        for line in inputfile:

            # Update multiline
            line = line.rstrip()  # pylint: disable=redefined-loop-name
            multiline += line
            multiline_len += 1

            # Skip empty lines
            if len(line) == 0 or multiline_len > ROUTE_MAX_MULTILINE_LEN:
                if len(line) > 0:
                    print('?', multiline)
                route_record_prefix = None
                route_record_proto = None
                multiline = ''
                multiline_len = 0
                continue

            # Parse multiline entry (with or without prefix)
            entry = unpack_lgl_entry(multiline)
            if entry is not None:

                # Extract route details
                prefix, nexthop_ip, med, locpref, aspath, origin = entry

                # Reset parser state
                multiline = ''
                multiline_len = 0

                ##############
                # BGP PREFIX #
                ##############

                # Extract prefix
                if prefix != '':

                    # Reset route record
                    route_record_prefix = None
                    route_record_proto = None

                    # Prepare prefix
                    prefix_int, mask = 0, 0
                    proto = IPV6 if ':' in prefix else IPV4
                    try:
                        # Ignore shortened prefix notation for /8 and /16
                        if prefix.endswith('.0.0') is True:
                            raise ValueError(f'indecisive classful {IPV4_STR} prefix ({prefix})')

                        # Fix shortened prefix notation for /24
                        if prefix.endswith('.0') is True:
                            prefix += '/24'

                        # Parse prefix
                        prefix_str, mask = prefix.split('/', 1)
                        if proto == IPV6:
                            net, host = struct_unpack(STRUCT_8B8B, socket_inet_pton(AF_INET6, prefix_str))
                            prefix_int = (net << 64) + host
                        else:
                            prefix_int = struct_unpack(STRUCT_4B, socket_inet_pton(AF_INET, prefix_str))[0]

                        # Sanitize prefix
                        max_mask, mask = (128 if proto == IPV6 else 32), int(mask)
                        shift = max_mask - mask
                        if mask > max_mask:
                            raise ValueError(f'invalid {IPV6_STR if proto == IPV6 else IPV4_STR} mask (/{mask})')
                        if (prefix_int >> shift) << shift != prefix_int:
                            raise ValueError(f'misaligned prefix ({prefix})')

                    # Handler errors
                    except ValueError as exc:

                        # Yield (or re-raise) route error
                        yield from route_error(FtlLglDataError('Unable to decode prefix',
                                               reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PREFIX, data=line, exception=exc))

                        # Reset parser state
                        multiline = ''
                        multiline_len = 0
                        continue

                    # Clone route record (filled with header data)
                    route_record_prefix = list(route_record_header)
                    route_record_proto = proto

                    # Update prefix count
                    if prefix != last_prefix:
                        n_prefixes += 1
                    last_prefix = prefix

                    # Add sequence to route record prefix
                    if FTL_ATTR_BGP_ROUTE_SEQUENCE >= 0:
                        route_record_prefix[FTL_ATTR_BGP_ROUTE_SEQUENCE] = n_prefixes

                    # Add peer protocol to route record prefix
                    if FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL >= 0:
                        peer_proto = proto
                        if FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN:
                            peer_proto = IPV6_STR if peer_proto == IPV6 else IPV4_STR
                        route_record_prefix[FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL] = peer_proto

                    # Add prefix protocol to route record prefix
                    if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL >= 0:
                        prefix_proto = proto
                        if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL_HUMAN:
                            prefix_proto = IPV6_STR if prefix_proto == IPV6 else IPV4_STR
                        route_record_prefix[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = prefix_proto

                    # Add prefix to route record prefix
                    if not FTL_ATTR_BGP_ROUTE_PREFIX_HUMAN:
                        prefix = (prefix_int, mask)
                    if FTL_ATTR_BGP_ROUTE_PREFIX >= 0:
                        route_record_prefix[FTL_ATTR_BGP_ROUTE_PREFIX] = prefix

                    # Add peer protocol to peer table record and yield on first prefix
                    # NOTE: We assume that all routes in a given looking glass dump share the same peer protocol
                    if FTL_RECORD_BGP_PEER_TABLE:
                        if peer_table_record is not None:
                            if FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL >= 0:
                                peer_proto = proto
                                if FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL_HUMAN:
                                    peer_proto = IPV6_STR if peer_proto == IPV6 else IPV4_STR
                                peer_table_record[FTL_ATTR_BGP_PEER_TABLE_PEER_PROTOCOL] = peer_proto

                            # Yield and nullify final peer table record
                            yield peer_table_emit(peer_table_record)
                            peer_table_record = None

                # Should not happen (only on errors or table start with missing first line)
                if route_record_prefix is None:
                    continue

                #############
                # BGP ROUTE #
                #############

                # Clone route record (filled with header+prefix data)
                route_record = list(route_record_prefix)

                # Extract nexthop protocol and IP
                try:
                    nexthop_ip_int = None
                    nexthop_proto = IPV6 if ':' in nexthop_ip else IPV4
                    if nexthop_proto == IPV6:
                        net, host = struct_unpack(STRUCT_8B8B, socket_inet_pton(AF_INET6, nexthop_ip))
                        nexthop_ip_int = (net << 64) + host
                    else:
                        nexthop_ip_int = struct_unpack(STRUCT_4B, socket_inet_pton(AF_INET, nexthop_ip))[0]
                    if FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL >= 0:
                        if FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL_HUMAN:
                            nexthop_proto = IPV6_STR if nexthop_proto == IPV6 else IPV4_STR
                        route_record[FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL] = nexthop_proto
                    if FTL_ATTR_BGP_ROUTE_NEXTHOP_IP >= 0:
                        if not FTL_ATTR_BGP_ROUTE_NEXTHOP_IP_HUMAN:
                            nexthop_ip = nexthop_ip_int
                        route_record[FTL_ATTR_BGP_ROUTE_NEXTHOP_IP] = nexthop_ip

                # Handle error
                except ValueError as exc:

                    # Yield (or re-raise) route error
                    yield from route_error(FtlLglDataError('Unable to decode nexthop IP',
                                           reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_IP, data=line, exception=exc))

                    # Reset route record and parser state
                    route_record_prefix = None
                    route_record_proto = None
                    multiline = ''
                    multiline_len = 0
                    continue

                # Extract MED metric
                try:
                    if med != '':
                        med_int = int(med)

                        # Add MED metric to route record
                        if FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC >= 0:
                            route_record[FTL_ATTR_BGP_ROUTE_MULTI_EXIT_DISC] = med_int

                # Handle error
                except ValueError as exc:

                    # Yield (or re-raise) route error
                    yield from route_error(FtlLglDataError('Unable to decode MED metric',
                                           reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR, data=line, exception=exc))

                    # Reset route record and parser state
                    route_record_prefix = None
                    route_record_proto = None
                    multiline = ''
                    multiline_len = 0
                    continue

                # Extract local pref
                try:
                    if locpref != '':
                        locpref_int = int(locpref)

                        # Add local_pref to route record
                        if FTL_ATTR_BGP_ROUTE_LOCAL_PREF >= 0:
                            route_record[FTL_ATTR_BGP_ROUTE_LOCAL_PREF] = locpref_int

                # Handle error
                except ValueError as exc:

                    # Yield (or re-raise) route error
                    yield from route_error(FtlLglDataError('Unable to decode local-pref',
                                           reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_ATTR, data=line, exception=exc))

                    # Reset route record and parser state
                    route_record_prefix = None
                    route_record_proto = None
                    multiline = ''
                    multiline_len = 0
                    continue

                # Extract AS path
                try:
                    aspath = (tuple() if aspath == '' else
                              tuple(tuple(sorted(set(int(y) for y in x[1:-1].split(',') if len(y) > 0)))
                                    if x[0] == '{' else int(x) for x in aspath.split()))

                    # Add AS path to route record
                    if FTL_ATTR_BGP_ROUTE_ASPATH >= 0:
                        route_record[FTL_ATTR_BGP_ROUTE_ASPATH] = aspath

                # Handle error
                except ValueError as exc:

                    # Yield (or re-raise) route error
                    yield from route_error(FtlLglDataError('Unable to decode AS path',
                                           reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_ASPATH, data=line, exception=exc))

                    # Reset route record and parser state
                    route_record_prefix = None
                    route_record_proto = None
                    multiline = ''
                    multiline_len = 0
                    continue

                # Add origin to route record
                if FTL_ATTR_BGP_ROUTE_ORIGIN >= 0:
                    origin_value = None
                    if origin == 'i':
                        origin_value = (FTL_ATTR_BGP_ROUTE_ORIGIN_IGP_STR if FTL_ATTR_BGP_ROUTE_ORIGIN_HUMAN else
                                        FTL_ATTR_BGP_ROUTE_ORIGIN_IGP)
                    elif origin == 'e':
                        origin_value = (FTL_ATTR_BGP_ROUTE_ORIGIN_EGP_STR if FTL_ATTR_BGP_ROUTE_ORIGIN_HUMAN else
                                        FTL_ATTR_BGP_ROUTE_ORIGIN_EGP)
                    elif origin == '?':
                        origin_value = (FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE_STR if FTL_ATTR_BGP_ROUTE_ORIGIN_HUMAN else
                                        FTL_ATTR_BGP_ROUTE_ORIGIN_INCOMPLETE)
                    if origin_value is not None:
                        route_record[FTL_ATTR_BGP_ROUTE_ORIGIN] = origin_value

                # Update stats
                if route_record_proto == IPV6:
                    n_routes6 += 1
                else:
                    n_routes4 += 1

                # Yield final route
                if FTL_RECORD_BGP_ROUTE:
                    yield route_emit(route_record)

        #############
        # BGP STATS #
        #############

        #  ------------------------
        #  [PCH] % sh bgp ipv4 wide
        #  ------------------------
        #  Displayed  612636 routes and 1443316 total paths

        #  ------------------------
        #  [PCH] % sh bgp ipv4 wide
        #  ------------------------
        #  Total number of prefixes 612636

        # Extract totals from last line for sanitarization
        match = ROUTE_TOTAL1_REGEX.match(multiline)
        if match is not None:
            try:
                prefixes = int(match.group(1))
                routes = int(match.group(2))
            except ValueError:
                pass
        else:
            match = ROUTE_TOTAL2_REGEX.match(multiline)
            if match is not None:
                try:
                    prefixes = int(match.group(1))
                except ValueError:
                    pass

        # Check prefix parsing result
        if prefixes is not None and prefixes != n_prefixes:
            message = 'Mismatching number of looking glass prefixes ({:,}/{:,})'.format(n_prefixes, prefixes)
            yield from bgp_error(FtlLglDataError(message, reason=FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA, data=line))

        # Check route parsing result
        if routes is not None and routes != n_routes4 + n_routes6:
            message = 'Mismatching number of looking glass routes ({:,}/{:,})'.format(n_routes4 + n_routes6, routes)
            yield from bgp_error(FtlLglDataError(message, reason=FTL_ATTR_BGP_ERROR_REASON_MISSING_DATA, data=line))

        # Update stats record
        if FTL_RECORD_BGP_STATS:

            # Add number of entries
            if FTL_ATTR_BGP_STATS_LGL_ENTRIES >= 0:
                stats_record[FTL_ATTR_BGP_STATS_LGL_ENTRIES] += n_prefixes

            # Add number of RIB routes
            if FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4 >= 0:
                stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV4] += n_routes4

            # Add number of RIB routes
            if FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6 >= 0:
                stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_RIB_IPV6] += n_routes6

    # Re-raise already handled header exceptions as format exceptions
    except FtlLglHeaderError as error:
        raise FtlLglFormatError(error=error)  # pylint: disable=raise-missing-from

    # Re-raise already handled record errors
    except FtlLglError as error:
        raise error

    # Yield (or re-raise) unhandled exceptions
    except Exception as exc:  # pylint: disable=broad-exception-caught
        yield from bgp_error(FtlLglDataError('Unhandled data error', data=line, trace=True, exception=exc))



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import sys
import traceback

# Local imports
from .const import FTL_ATTR_BGP_ERROR_SOURCE_MRT
from .const import FTL_ATTR_BGP_ERROR_SOURCE_LGL
from .const import FTL_ATTR_BGP_ERROR_SOURCE_PARSER
from .const import FTL_ATTR_BGP_ERROR_SCOPE_BASE
from .const import FTL_ATTR_BGP_ERROR_SCOPE_FILE
from .const import FTL_ATTR_BGP_ERROR_SCOPE_HEADER
from .const import FTL_ATTR_BGP_ERROR_SCOPE_FORMAT
from .const import FTL_ATTR_BGP_ERROR_SCOPE_DATA
from .const import FTL_ATTR_BGP_ERROR_REASON_RUNTIME


##################
# GENERIC ERRORS #
##################

class FtlError(Exception):
    """ Generic exception.
    """
    # Prepare internals
    source = FTL_ATTR_BGP_ERROR_SOURCE_PARSER
    scope = FTL_ATTR_BGP_ERROR_SCOPE_BASE

    def __init__(self, message=None, reason=FTL_ATTR_BGP_ERROR_REASON_RUNTIME, record=None, data=None, trace=None,
                 error=None, exception=None, **_):
        """ Initialize exception instance.
        """
        # Prepare error details
        self.message = message
        self.reason = reason
        self.record = record
        self.data = data
        self.trace = trace

        # Extract stack trace
        if trace is True:
            tbline = 'Traceback (most recent call last):'
            self.trace, (cls, exc, tb) = list(), sys.exc_info()
            for line in [tbline] + traceback.format_tb(tb) + traceback.format_exception_only(cls, exc):
                for err in line.rstrip().split('\n'):
                    errline, indent = err.lstrip(), len(err)
                    self.trace.append((errline, indent - len(errline)))

        # Update with parent error
        if error is not None:
            self.__cause__ = None
            self.__traceback__ = error.__traceback__
            if self.source == FTL_ATTR_BGP_ERROR_SOURCE_PARSER:
                self.source = error.source
            if self.reason == FTL_ATTR_BGP_ERROR_SCOPE_BASE:
                self.scope = error.scope
            if self.reason == FTL_ATTR_BGP_ERROR_REASON_RUNTIME:
                self.reason = error.reason
            if self.message is None:
                self.message = error.message
            if self.record is None:
                self.record = record
            if self.data is None:
                self.data = error.data
            if self.trace is None:
                self.trace = error.trace

        # Update with parent exception
        if exception is not None:
            self.__cause__ = None
            self.__traceback__ = exception.__traceback__
            excmsg = f'[{exception.__class__.__name__}] {str(exception)}'
            self.message = f'{self.message}: {excmsg}' if self.message is not None else excmsg

        # Invoke super constructor
        errmsg = f'ERR|{self.source}.{self.scope}.{self.reason}'
        errmsg = f'[{errmsg}]' if self.record is None else f'[{errmsg}-{self.record}]'.upper()
        if self.message is not None:
            errmsg = f'{errmsg} {self.message}'
        super().__init__(errmsg)


class FtlFileError(FtlError):
    """ Generic file exception.
    """
    # Update internals
    scope = FTL_ATTR_BGP_ERROR_SCOPE_FILE


class FtlFormatError(FtlError):
    """ Generic file format exception.
    """
    # Update internals
    scope = FTL_ATTR_BGP_ERROR_SCOPE_FORMAT


class FtlDataError(FtlError):
    """ Generic file data exception.
    """
    # Update internals
    scope = FTL_ATTR_BGP_ERROR_SCOPE_DATA


##############
# MRT ERRORS #
##############

class FtlMrtError(FtlError):
    """ Base MRT exception.
    """
    # Update internals
    source = FTL_ATTR_BGP_ERROR_SOURCE_MRT


class FtlMrtHeaderError(FtlMrtError):
    """ Exception for invalid MRT header.
    """
    # Update internals
    scope = FTL_ATTR_BGP_ERROR_SCOPE_HEADER


class FtlMrtFormatError(FtlMrtError, FtlFormatError):
    """ Exception for invalid MRT format.
    """


class FtlMrtDataError(FtlMrtError, FtlDataError):
    """ Exception for invalid MRT data.
    """


##############
# LGL ERRORS #
##############

class FtlLglError(FtlError):
    """ Base looking glass exception.
    """
    # Update internals
    source = FTL_ATTR_BGP_ERROR_SOURCE_LGL


class FtlLglHeaderError(FtlLglError):
    """ Exception for invalid looking glass header.
    """
    # Update internals
    scope = FTL_ATTR_BGP_ERROR_SCOPE_HEADER


class FtlLglFormatError(FtlLglError, FtlFormatError):
    """ Exception for invalid looking glass format.
    """


class FtlLglDataError(FtlLglError, FtlDataError):
    """ Exception for invalid looking glass data.
    """



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import argparse
import sys

# Local imports
from . import __version__
from . import BgpParser
from . import FtlError
from . import FtlDataError


def main():
    """ Main routine.
    """
    # Prepare command line arguments
    description = f'ftlbgp [v{__version__}] - Parse BGP data in MRT or looking glass format'
    parser = argparse.ArgumentParser(prog='ftlbgp', description=description)
    argparse._VersionAction.__call__ = lambda *_: (help(BgpParser), parser.exit())  # pylint: disable=protected-access
    parser.add_argument('filelist', metavar='<FILE>', nargs='+', help='input file with BGP data (supports bz2/gz)')
    parser.add_argument('--pkg-help', action='version', help='show package help for BgpParser usage')
    parser.add_argument('--json', default=False, action='store_true', help='output BGP records in JSON format')
    parser.format_help = lambda: argparse.ArgumentParser.format_help(parser) + '\n'
    parser._get_formatter = lambda: parser.formatter_class(prog='python3 -m ftlbgp')  # pylint: disable=protected-access
    args = parser.parse_args()

    # Initialize BGP parser
    with BgpParser(named_records=args.json, serialize=True, raise_on_errors=False,
                   bgp_records=BgpParser.bgp.records.route,
                   bgp_route=(
                       BgpParser.bgp.route.human.DEFAULT
                       & ~BgpParser.bgp.route.human.type
                       & ~BgpParser.bgp.route.human.peer_bgp_id
                   )) as parse:

        # Parse input files
        for filename in args.filelist:
            try:
                # Iterate and print records
                for record in parse(filename):
                    print(record)

            # Ignore non-critical data errors
            except FtlDataError:
                continue

            # Abort on unknown errors
            except FtlError as error:
                print(str(error))
                sys.exit(1)

            # Handle command line induced errors
            except (BrokenPipeError, KeyboardInterrupt):
                return


# Main routine
if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# pylint: disable=I0011,I0020,I0021,I0023
# pylint: disable=W0149,W0717,R0902,R0904,R0911,R0912,R0913,R0914,R0915,R1702,R1734,R1735,R2044,R6103,C0103,C0209,C2001
# pylint: enable=I0011
""" ftlbgp
Copyright (C) 2014-2025 Leitwert GmbH

This software is distributed under the terms of the MIT license.
It can be found in the LICENSE file or at https://opensource.org/licenses/MIT.

Author Johann SCHLAMP <schlamp@leitwert.net>
"""

# System imports
import base64

# Local imports
from .const import BGP_SAFI_UNICAST
from .const import BGP_SAFI_MULTICAST
from .const import BGP_BGP4MP_UPDATE
from .const import BGP_BGP4MP_KEEPALIVE
from .const import BGP_BGP4MP_ROUTE_REFRESH
from .const import BGP_BGP4MP_NOTIFICATION
from .const import BGP_BGP4MP_OPEN
from .const import BGP_PARAMS_CAPABILITIES
from .nlri import unpack_mrt_bgp_nlri
from .attr import unpack_mrt_bgp_attr
from ...util import CACHE_TS
from ...const import IPV4
from ...const import IPV6
from ...const import IPV4_STR
from ...const import IPV6_STR
from ...const import AF_INET
from ...const import AF_INET6
from ...const import AFI_IPV4
from ...const import AFI_IPV6
from ...const import STRUCT_2B
from ...const import STRUCT_4B
from ...const import STRUCT_8B8B
from ...const import DATETIME_FORMAT_USEC
from ...const import DATETIME_FORMAT_MIN
from ...const import struct_unpack
from ...const import socket_inet_ntop
from ...const import datetime_utcfromtimestamp
from ....model.attr import FTL_ATTR_BGP_ROUTE_SOURCE
from ....model.attr import FTL_ATTR_BGP_ROUTE_SEQUENCE
from ....model.attr import FTL_ATTR_BGP_ROUTE_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_AS
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_IP
from ....model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_NEXTHOP_IP
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_PREFIX
from ....model.attr import FTL_ATTR_BGP_ROUTE_PATH_ID
from ....model.attr import FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_KEEP_ALIVE_PEER_AS
from ....model.attr import FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP
from ....model.attr import FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_PEER_AS
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_PEER_AS
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_PEER_IP
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_DATA
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE_HUMAN
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE_HUMAN
from ....model.attr import FTL_ATTR_BGP_NOTIFICATION_DATA_HUMAN
from ....model.attr import FTL_ATTR_BGP_OPEN_TIMESTAMP
from ....model.attr import FTL_ATTR_BGP_OPEN_PEER_PROTOCOL
from ....model.attr import FTL_ATTR_BGP_OPEN_PEER_AS
from ....model.attr import FTL_ATTR_BGP_OPEN_PEER_IP
from ....model.attr import FTL_ATTR_BGP_OPEN_VERSION
from ....model.attr import FTL_ATTR_BGP_OPEN_MY_AS
from ....model.attr import FTL_ATTR_BGP_OPEN_HOLD_TIME
from ....model.attr import FTL_ATTR_BGP_OPEN_BGP_ID
from ....model.attr import FTL_ATTR_BGP_OPEN_CAPABILITIES
from ....model.attr import FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN
from ....model.attr import FTL_ATTR_BGP_OPEN_TIMESTAMP_HUMAN
from ....model.attr import FTL_ATTR_BGP_OPEN_PEER_PROTOCOL_HUMAN
from ....model.attr import FTL_ATTR_BGP_OPEN_PEER_IP_HUMAN
from ....model.attr import FTL_ATTR_BGP_OPEN_BGP_ID_HUMAN
from ....model.attr import FTL_ATTR_BGP_OPEN_CAPABILITIES_HUMAN
from ....model.attr import FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN_HUMAN
from ....model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV4
from ....model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV6
from ....model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV4
from ....model.attr import FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV6
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES
from ....model.attr import FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES_HUMAN
from ....model.const import FTL_ATTR_BGP_ROUTE_SOURCE_ANNOUNCE
from ....model.const import FTL_ATTR_BGP_ROUTE_SOURCE_ANNOUNCE_STR
from ....model.const import FTL_ATTR_BGP_ROUTE_SOURCE_WITHDRAW
from ....model.const import FTL_ATTR_BGP_ROUTE_SOURCE_WITHDRAW_STR
from ....model.const import FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE_TO_STR
from ....model.const import FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE_TO_STR
from ....model.const import FTL_ATTR_BGP_OPEN_CAPABILITY_TO_STR
from ....model.const import FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL
from ....model.record import FTL_RECORD_BGP_ROUTE
from ....model.record import FTL_RECORD_BGP_KEEP_ALIVE
from ....model.record import FTL_RECORD_BGP_ROUTE_REFRESH
from ....model.record import FTL_RECORD_BGP_NOTIFICATION
from ....model.record import FTL_RECORD_BGP_OPEN
from ....model.record import FTL_RECORD_BGP_STATS
from ....model.error import FtlMrtDataError


def unpack_mrt_bgp_msg(caches, stats_record, bgp_error, route_records, keep_alive_records, route_refresh_records,
                       notification_records, open_records, msg_bytes, mtype, sequence, ts, peer_as, peer_ip, peer_afinet,
                       aslen=4, addpath=False):
    """ Parse MRT BGP/BGP4MP message.
    """
    # Prepare byte offset
    offset = 0

    ######################
    # BGP UPDATE MESSAGE #
    ######################

    # ------------------------------------
    # [RFC4271] 4.3. UPDATE Message Format
    # ------------------------------------
    # +-----------------------------------------------------+
    # |   Withdrawn Routes Length (2 octets)                |
    # +-----------------------------------------------------+
    # |   Withdrawn Routes (variable)                       |
    # +-----------------------------------------------------+
    # |   Total Path Attribute Length (2 octets)            |
    # +-----------------------------------------------------+
    # |   Path Attributes (variable)                        |
    # +-----------------------------------------------------+
    # |   Network Layer Reachability Information (variable) |
    # +-----------------------------------------------------+

    # Parse UPDATE message
    if mtype == BGP_BGP4MP_UPDATE:

        # Access route record template
        route_init, route_emit, route_error = route_records

        # Initialize route record reach
        route_record_reach = list(route_init)

        # Add source to route record reach
        if FTL_ATTR_BGP_ROUTE_SOURCE >= 0:
            if FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN:
                route_record_reach[FTL_ATTR_BGP_ROUTE_SOURCE] = FTL_ATTR_BGP_ROUTE_SOURCE_ANNOUNCE_STR
            else:
                route_record_reach[FTL_ATTR_BGP_ROUTE_SOURCE] = FTL_ATTR_BGP_ROUTE_SOURCE_ANNOUNCE

        # Add sequence number to route record reach
        if FTL_ATTR_BGP_ROUTE_SEQUENCE >= 0:
            route_record_reach[FTL_ATTR_BGP_ROUTE_SEQUENCE] = sequence

        # Add timestamp to route record reach
        if FTL_ATTR_BGP_ROUTE_TIMESTAMP >= 0:
            ts_route = ts

            # Cache date string for minute-based timestamp
            if FTL_ATTR_BGP_ROUTE_TIMESTAMP_HUMAN:
                if caches:
                    ts_cache = caches[CACHE_TS]
                    ts_min, ts_sec = divmod(ts_route, 60)
                    ts_cached = ts_cache.get(ts_min, None)
                    if ts_cached is None:
                        ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                        ts_cache[ts_min] = ts_cached

                    # Add seconds and microseconds
                    ts_route = f'{ts_cached}:{ts_sec:09.6f}'

                # Do not use cache
                else:
                    ts_route = datetime_utcfromtimestamp(ts_route).strftime(DATETIME_FORMAT_USEC)

            # Add timestamp
            route_record_reach[FTL_ATTR_BGP_ROUTE_TIMESTAMP] = ts_route

        # Add peer protocol to route record reach
        if FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL >= 0:
            if FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL_HUMAN:
                route_record_reach[FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL] = IPV6_STR if peer_afinet == AF_INET6 else IPV4_STR
            else:
                route_record_reach[FTL_ATTR_BGP_ROUTE_PEER_PROTOCOL] = IPV6 if peer_afinet == AF_INET6 else IPV4

        # Add peer AS to route record reach
        if FTL_ATTR_BGP_ROUTE_PEER_AS >= 0:
            route_record_reach[FTL_ATTR_BGP_ROUTE_PEER_AS] = peer_as

        # Add peer IP to route record reach
        if FTL_ATTR_BGP_ROUTE_PEER_IP >= 0:
            peer_ip_route = peer_ip
            if FTL_ATTR_BGP_ROUTE_PEER_IP_HUMAN:
                peer_ip_route = socket_inet_ntop(peer_afinet, peer_ip_route)
            elif peer_afinet == AF_INET6:
                net, host = struct_unpack(STRUCT_8B8B, peer_ip_route)
                peer_ip_route = (net << 64) + host
            else:
                peer_ip_route = struct_unpack(STRUCT_4B, peer_ip_route)[0]
            route_record_reach[FTL_ATTR_BGP_ROUTE_PEER_IP] = peer_ip_route

        # Initialize route record unreach
        route_record_unreach = list(route_record_reach)

        # Add source to route record unreach
        if FTL_ATTR_BGP_ROUTE_SOURCE >= 0:
            if FTL_ATTR_BGP_ROUTE_SOURCE_HUMAN:
                route_record_unreach[FTL_ATTR_BGP_ROUTE_SOURCE] = FTL_ATTR_BGP_ROUTE_SOURCE_WITHDRAW_STR
            else:
                route_record_unreach[FTL_ATTR_BGP_ROUTE_SOURCE] = FTL_ATTR_BGP_ROUTE_SOURCE_WITHDRAW

        # Initialize announced/withdrawn prefixes
        nlri, nlui = tuple(), tuple()

        # Parse UPDATE message data
        try:

            # Parse withdrawn routes length
            ulen = struct_unpack(STRUCT_2B, msg_bytes[offset:offset + 2])[0]
            offset += 2

            # Parse NLUI
            if ulen > 0:
                nlui = unpack_mrt_bgp_nlri(caches, stats_record, msg_bytes[offset:offset + ulen], AF_INET,
                                           addpath=addpath)
                offset += ulen

            # Parse attributes length
            alen = struct_unpack(STRUCT_2B, msg_bytes[offset:offset + 2])[0]
            offset += 2

            # Parse attributes
            # NOTE: BGP path attributes are applied to announced routes only (not valid for withdrawn routes)
            mp_nlri, mp_nlui = unpack_mrt_bgp_attr(caches, stats_record, route_init, route_emit, route_record_reach,
                                                   msg_bytes[offset:offset + alen], aslen=aslen, addpath=addpath)
            offset += alen

            # Parse NLRI
            rlen = len(msg_bytes) - offset
            if rlen > 0:
                nlri = unpack_mrt_bgp_nlri(caches, stats_record, msg_bytes[offset:offset + rlen], AF_INET,
                                           addpath=addpath)
                offset += rlen

            # Update NLUI/NLRI
            nlui += mp_nlui
            nlri += mp_nlri

        # Yield (or re-raise) UPDATE message errors
        except FtlMrtDataError as error:
            yield from route_error(error)
            return

        # -----------------------------------
        # [RFC4271] 4.3 UPDATE Message Format
        # -----------------------------------
        #
        # NOTE: The following procedure should be implemented by clients
        #
        # An UPDATE message SHOULD NOT include the same address prefix in the
        # WITHDRAWN ROUTES and Network Layer Reachability Information fields.
        # However, a BGP speaker MUST be able to process UPDATE messages in
        # this form. A BGP speaker SHOULD treat an UPDATE message of this form
        # as though the WITHDRAWN ROUTES do not contain the address prefix.

        # Finalize route records unreach (withdrawls)
        # NOTE: Human-readable conversions are done in unpack_mrt_bgp_attr() and unpack_mrt_bgp_nlri()
        for afinet, prefix_proto, prefix, path_id, _, _ in nlui:

            # Add prefix protocol to route record unreach
            if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL >= 0:
                route_record_unreach[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = prefix_proto

            # Add prefix to route record unreach
            if FTL_ATTR_BGP_ROUTE_PREFIX >= 0:
                route_record_unreach[FTL_ATTR_BGP_ROUTE_PREFIX] = prefix

            # Add ADD-PATH path ID to route record unreach
            if FTL_ATTR_BGP_ROUTE_PATH_ID >= 0:
                route_record_unreach[FTL_ATTR_BGP_ROUTE_PATH_ID] = path_id

            # Update stats record
            if FTL_RECORD_BGP_STATS:

                # Update IPv4 routes announced
                if afinet == AF_INET:
                    if FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV4 >= 0:
                        stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV4] += 1

                # Update IPv6 routes announced
                elif afinet == AF_INET6:  # pylint: disable=confusing-consecutive-elif
                    if FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV6 >= 0:
                        stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_WITHDRAW_IPV6] += 1

            # Yield final route record unreach
            if FTL_RECORD_BGP_ROUTE:
                yield route_emit(route_record_unreach)

        # Finalize route records reach (announcements)
        # NOTE: Human-readable conversions are done in unpack_mrt_bgp_attr() and unpack_mrt_bgp_nlri()
        for afinet, prefix_proto, prefix, path_id, nexthop_proto, nexthop_ip in nlri:

            # Initialize route record reach_mp
            # NOTE: Route records might contain a pre-MP NEXT_HOP attribute that applies to all pre-MP NLRIs,
            # NOTE: so records must be re-initialized for MP NLRI entries
            route_record_reach_mp = route_record_reach

            # Add nexthop protocol to route record reach_mp
            if FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL >= 0 and nexthop_proto is not None:
                route_record_reach_mp = list(route_record_reach)
                route_record_reach_mp[FTL_ATTR_BGP_ROUTE_NEXTHOP_PROTOCOL] = nexthop_proto

                # Add nexthop IP to route record reach_mp
                if FTL_ATTR_BGP_ROUTE_NEXTHOP_IP >= 0:
                    route_record_reach_mp[FTL_ATTR_BGP_ROUTE_NEXTHOP_IP] = nexthop_ip

            # Add prefix protocol to route record reach_mp
            if FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL >= 0:
                route_record_reach_mp[FTL_ATTR_BGP_ROUTE_PREFIX_PROTOCOL] = prefix_proto

            # Add prefix to route record reach_mp
            if FTL_ATTR_BGP_ROUTE_PREFIX >= 0:
                route_record_reach_mp[FTL_ATTR_BGP_ROUTE_PREFIX] = prefix

            # Add ADD-PATH path ID to route record reach_mp
            if FTL_ATTR_BGP_ROUTE_PATH_ID >= 0:
                route_record_reach_mp[FTL_ATTR_BGP_ROUTE_PATH_ID] = path_id

            # Update stats record
            if FTL_RECORD_BGP_STATS:

                # Update IPv4 routes withdrawn
                if afinet == AF_INET:
                    if FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV4 >= 0:
                        stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV4] += 1

                # Update IPv6 routes withdrawn
                elif afinet == AF_INET6:  # pylint: disable=confusing-consecutive-elif
                    if FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV6 >= 0:
                        stats_record[FTL_ATTR_BGP_STATS_BGP_ROUTES_ANNOUNCE_IPV6] += 1

            # Yield final route record reach_mp
            if FTL_RECORD_BGP_ROUTE:
                yield route_emit(route_record_reach_mp)

    #########################
    # BGP KEEPALIVE MESSAGE #
    #########################

    # ---------------------------------------
    # [RFC4271] 4.4. KEEPALIVE Message Format
    # ---------------------------------------
    # A KEEPALIVE message consists of only the message header and has a
    # length of 19 octets.

    # Parse KEEPALIVE message
    elif mtype == BGP_BGP4MP_KEEPALIVE:

        # Access keep alive record template
        keep_alive_init, keep_alive_emit, _ = keep_alive_records

        # Initialize keep alive record
        keep_alive_record = list(keep_alive_init)

        # Add timestamp to keep alive record
        if FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP >= 0:
            ts_keep_alive = ts

            # Cache date string for minute-based timestamp
            if FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP_HUMAN:
                if caches:
                    ts_cache = caches[CACHE_TS]
                    ts_min, ts_sec = divmod(ts_keep_alive, 60)
                    ts_cached = ts_cache.get(ts_min, None)
                    if ts_cached is None:
                        ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                        ts_cache[ts_min] = ts_cached

                    # Add seconds and microseconds
                    ts_keep_alive = f'{ts_cached}:{ts_sec:09.6f}'

                # Do not use cache
                else:
                    ts_keep_alive = datetime_utcfromtimestamp(ts_keep_alive).strftime(DATETIME_FORMAT_USEC)

            # Add timestamp
            keep_alive_record[FTL_ATTR_BGP_KEEP_ALIVE_TIMESTAMP] = ts_keep_alive

        # Add peer protocol to keep alive record
        if FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL >= 0:
            if FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL_HUMAN:
                keep_alive_record[FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL] = (IPV6_STR if peer_afinet == AF_INET6
                                                                            else IPV4_STR)
            else:
                keep_alive_record[FTL_ATTR_BGP_KEEP_ALIVE_PEER_PROTOCOL] = IPV6 if peer_afinet == AF_INET6 else IPV4

        # Add peer AS to keep alive record
        if FTL_ATTR_BGP_KEEP_ALIVE_PEER_AS >= 0:
            keep_alive_record[FTL_ATTR_BGP_KEEP_ALIVE_PEER_AS] = peer_as

        # Add peer IP to keep alive record
        if FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP >= 0:
            peer_ip_keep_alive = peer_ip
            if FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP_HUMAN:
                peer_ip_keep_alive = socket_inet_ntop(peer_afinet, peer_ip_keep_alive)
            elif peer_afinet == AF_INET6:
                net, host = struct_unpack(STRUCT_8B8B, peer_ip_keep_alive)
                peer_ip_keep_alive = (net << 64) + host
            else:
                peer_ip_keep_alive = struct_unpack(STRUCT_4B, peer_ip_keep_alive)[0]
            keep_alive_record[FTL_ATTR_BGP_KEEP_ALIVE_PEER_IP] = peer_ip_keep_alive

        # Yield final keep alive record
        if FTL_RECORD_BGP_KEEP_ALIVE:
            yield keep_alive_emit(keep_alive_record)

    #############################
    # BGP ROUTE_REFRESH MESSAGE #
    #############################

    # ----------------------------------
    # [RFC2918] 3. Route-REFRESH Message
    # ----------------------------------
    # 0       7      15      23      31
    # +-------+-------+-------+-------+
    # |      AFI      | Res.  | SAFI  |
    # +-------+-------+-------+-------+

    # Parse ROUTE_REFRESH message
    elif mtype == BGP_BGP4MP_ROUTE_REFRESH:  # pylint: disable=confusing-consecutive-elif

        # Access route refresh template
        route_refresh_init, route_refresh_emit, _ = route_refresh_records

        # Initialize route refresh record
        route_refresh_record = list(route_refresh_init)

        # Add timestamp to route refresh record
        if FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP >= 0:
            ts_route_refresh = ts

            # Cache date string for minute-based timestamp
            if FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP_HUMAN:
                if caches:
                    ts_cache = caches[CACHE_TS]
                    ts_min, ts_sec = divmod(ts_route_refresh, 60)
                    ts_cached = ts_cache.get(ts_min, None)
                    if ts_cached is None:
                        ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                        ts_cache[ts_min] = ts_cached

                    # Add seconds and microseconds
                    ts_route_refresh = f'{ts_cached}:{ts_sec:09.6f}'

                # Do not use cache
                else:
                    ts_route_refresh = datetime_utcfromtimestamp(ts_route_refresh).strftime(DATETIME_FORMAT_USEC)

            # Add timestamp
            route_refresh_record[FTL_ATTR_BGP_ROUTE_REFRESH_TIMESTAMP] = ts_route_refresh

        # Add peer protocol to route refresh record
        if FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL >= 0:
            if FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL_HUMAN:
                route_refresh_record[FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL] = (IPV6_STR if peer_afinet == AF_INET6
                                                                                  else IPV4_STR)
            else:
                route_refresh_record[FTL_ATTR_BGP_ROUTE_REFRESH_PEER_PROTOCOL] = IPV6 if peer_afinet == AF_INET6 else IPV4

        # Add peer AS to route refresh record
        if FTL_ATTR_BGP_ROUTE_REFRESH_PEER_AS >= 0:
            route_refresh_record[FTL_ATTR_BGP_ROUTE_REFRESH_PEER_AS] = peer_as

        # Add peer IP to route refresh record
        if FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP >= 0:
            peer_ip_route_refresh = peer_ip
            if FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP_HUMAN:
                peer_ip_route_refresh = socket_inet_ntop(peer_afinet, peer_ip_route_refresh)
            elif peer_afinet == AF_INET6:
                net, host = struct_unpack(STRUCT_8B8B, peer_ip_route_refresh)
                peer_ip_route_refresh = (net << 64) + host
            else:
                peer_ip_route_refresh = struct_unpack(STRUCT_4B, peer_ip_route_refresh)[0]
            route_refresh_record[FTL_ATTR_BGP_ROUTE_REFRESH_PEER_IP] = peer_ip_route_refresh

        # Parse AFI value
        refresh_afi = struct_unpack(STRUCT_2B, msg_bytes[offset:offset + 2])[0]
        offset += 2

        # Skip reserved byte
        offset += 1

        # Parse SAFI value
        refresh_safi = msg_bytes[offset]

        # Check AFI value
        if refresh_afi != AFI_IPV4 and refresh_afi != AFI_IPV6:  # pylint: disable=consider-using-in
            yield from bgp_error(FtlMrtDataError(f'Invalid AFI value ({refresh_afi}) in BGP4MP route-refresh',
                                                 reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=msg_bytes))
            return

        # Check SAFI value
        if refresh_safi != BGP_SAFI_UNICAST and refresh_safi != BGP_SAFI_MULTICAST:  # pylint: disable=consider-using-in
            yield from bgp_error(FtlMrtDataError(f'Unsupported SAFI value ({refresh_safi}) in BGP4MP route-refresh',
                                                 reason=FTL_ATTR_BGP_ERROR_REASON_INVALID_PROTOCOL, data=msg_bytes))
            return

        # Parse protocol
        if FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL >= 0:
            refresh_proto = IPV6 if refresh_afi == AFI_IPV6 else IPV4
            if FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL_HUMAN:
                refresh_proto = IPV6_STR if refresh_afi == AFI_IPV6 else IPV4_STR
            route_refresh_record[FTL_ATTR_BGP_ROUTE_REFRESH_REFRESH_PROTOCOL] = refresh_proto

        # Yield final route refresh record
        if FTL_RECORD_BGP_ROUTE_REFRESH:
            yield route_refresh_emit(route_refresh_record)

    ############################
    # BGP NOTIFICATION MESSAGE #
    ############################

    # -------------------------------------
    # [RFC4271] NOTIFICATION Message Format
    # -------------------------------------
    # 0                   1                   2                   3
    # 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # | Error code    | Error subcode |   Data (variable)             |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    # Parse NOTIFICATION message
    elif mtype == BGP_BGP4MP_NOTIFICATION:  # pylint: disable=confusing-consecutive-elif

        # Access notification template
        notification_init, notification_emit, _ = notification_records

        # Initialize notification record
        notification_record = list(notification_init)

        # Add timestamp to notification record
        if FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP >= 0:
            ts_notification = ts

            # Cache date string for minute-based timestamp
            if FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP_HUMAN:
                if caches:
                    ts_cache = caches[CACHE_TS]
                    ts_min, ts_sec = divmod(ts_notification, 60)
                    ts_cached = ts_cache.get(ts_min, None)
                    if ts_cached is None:
                        ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                        ts_cache[ts_min] = ts_cached

                    # Add seconds and microseconds
                    ts_notification = f'{ts_cached}:{ts_sec:09.6f}'

                # Do not use cache
                else:
                    ts_notification = datetime_utcfromtimestamp(ts_notification).strftime(DATETIME_FORMAT_USEC)

            # Add timestamp
            notification_record[FTL_ATTR_BGP_NOTIFICATION_TIMESTAMP] = ts_notification

        # Add peer protocol to notification record
        if FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL >= 0:
            if FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL_HUMAN:
                notification_record[FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL] = (IPV6_STR if peer_afinet == AF_INET6
                                                                                else IPV4_STR)
            else:
                notification_record[FTL_ATTR_BGP_NOTIFICATION_PEER_PROTOCOL] = IPV6 if peer_afinet == AF_INET6 else IPV4

        # Add peer AS to notification record
        if FTL_ATTR_BGP_NOTIFICATION_PEER_AS >= 0:
            notification_record[FTL_ATTR_BGP_NOTIFICATION_PEER_AS] = peer_as

        # Add peer IP to notification record
        if FTL_ATTR_BGP_NOTIFICATION_PEER_IP >= 0:
            peer_ip_notification = peer_ip
            if FTL_ATTR_BGP_NOTIFICATION_PEER_IP_HUMAN:
                peer_ip_notification = socket_inet_ntop(peer_afinet, peer_ip_notification)
            elif peer_afinet == AF_INET6:
                net, host = struct_unpack(STRUCT_8B8B, peer_ip_notification)
                peer_ip_notification = (net << 64) + host
            else:
                peer_ip_notification = struct_unpack(STRUCT_4B, peer_ip_notification)[0]
            notification_record[FTL_ATTR_BGP_NOTIFICATION_PEER_IP] = peer_ip_notification

        # Parse error code
        err_code_int = msg_bytes[offset]
        if FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE >= 0:
            err_code = err_code_int
            if FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE_HUMAN:
                err_code = FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE_TO_STR.get(err_code, str(err_code))
            notification_record[FTL_ATTR_BGP_NOTIFICATION_ERROR_CODE] = err_code
        offset += 1

        # Parse error subcode
        if FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE >= 0:
            err_scode = msg_bytes[offset]
            if FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE_HUMAN:
                err_scode = FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE_TO_STR.get(err_code_int,
                                                                               dict()).get(err_scode, str(err_scode))
            notification_record[FTL_ATTR_BGP_NOTIFICATION_ERROR_SUBCODE] = err_scode
        offset += 1

        # Parse data
        if FTL_ATTR_BGP_NOTIFICATION_DATA >= 0:
            if offset < len(msg_bytes):
                data = msg_bytes[offset:]
                if FTL_ATTR_BGP_NOTIFICATION_DATA_HUMAN:
                    notification_record[FTL_ATTR_BGP_NOTIFICATION_DATA] = ' '.join('{:02x}'.format(byte) for byte in data)
                else:
                    notification_record[FTL_ATTR_BGP_NOTIFICATION_DATA] = base64.b64encode(data).decode('ascii')

        # Yield final notification record
        if FTL_RECORD_BGP_NOTIFICATION:
            yield notification_emit(notification_record)

    ####################
    # BGP OPEN MESSAGE #
    ####################

    # ----------------------------------
    # [RFC4271] 4.2. OPEN Message Format
    # ----------------------------------
    # 0                   1                   2                   3
    # 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    # +-+-+-+-+-+-+-+-+
    # |    Version    |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |     My Autonomous System      |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |           Hold Time           |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                         BGP Identifier                        |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # | Opt Parm Len  |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    # |                                                               |
    # |             Optional Parameters (variable)                    |
    # |                                                               |
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    #
    # 0                   1
    # 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-...
    # |  Parm. Type   | Parm. Length  |  Parameter Value (variable)
    # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-...
    #
    # Parameter Type is a one octet field that unambiguously
    # identifies individual parameters. Parameter Length is a one
    # octet field that contains the length of the Parameter Value
    # field in octets. Parameter Value is a variable length field
    # that is interpreted according to the value of the Parameter
    # Type field.

    # Parse OPEN message
    elif mtype == BGP_BGP4MP_OPEN:  # pylint: disable=confusing-consecutive-elif

        # Access open template
        open_init, open_emit, _ = open_records

        # Initialize open record
        open_record = list(open_init)

        # Add timestamp to open record
        if FTL_ATTR_BGP_OPEN_TIMESTAMP >= 0:
            ts_open = ts

            # Cache date string for minute-based timestamp
            if FTL_ATTR_BGP_OPEN_TIMESTAMP_HUMAN:
                if caches:
                    ts_cache = caches[CACHE_TS]
                    ts_min, ts_sec = divmod(ts_open, 60)
                    ts_cached = ts_cache.get(ts_min, None)
                    if ts_cached is None:
                        ts_cached = datetime_utcfromtimestamp(ts).strftime(DATETIME_FORMAT_MIN)
                        ts_cache[ts_min] = ts_cached

                    # Add seconds and microseconds
                    ts_open = f'{ts_cached}:{ts_sec:09.6f}'

                # Do not use cache
                else:
                    ts_open = datetime_utcfromtimestamp(ts_open).strftime(DATETIME_FORMAT_USEC)

            # Add timestamp
            open_record[FTL_ATTR_BGP_OPEN_TIMESTAMP] = ts_open

        # Add peer protocol to open record
        if FTL_ATTR_BGP_OPEN_PEER_PROTOCOL >= 0:
            if FTL_ATTR_BGP_OPEN_PEER_PROTOCOL_HUMAN:
                open_record[FTL_ATTR_BGP_OPEN_PEER_PROTOCOL] = IPV6_STR if peer_afinet == AF_INET6 else IPV4_STR
            else:
                open_record[FTL_ATTR_BGP_OPEN_PEER_PROTOCOL] = IPV6 if peer_afinet == AF_INET6 else IPV4

        # Add peer AS to open record
        if FTL_ATTR_BGP_OPEN_PEER_AS >= 0:
            open_record[FTL_ATTR_BGP_OPEN_PEER_AS] = peer_as

        # Add peer IP to open record
        if FTL_ATTR_BGP_OPEN_PEER_IP >= 0:
            peer_ip_open = peer_ip
            if FTL_ATTR_BGP_OPEN_PEER_IP_HUMAN:
                peer_ip_open = socket_inet_ntop(peer_afinet, peer_ip_open)
            elif peer_afinet == AF_INET6:
                net, host = struct_unpack(STRUCT_8B8B, peer_ip_open)
                peer_ip_open = (net << 64) + host
            else:
                peer_ip_open = struct_unpack(STRUCT_4B, peer_ip_open)[0]
            open_record[FTL_ATTR_BGP_OPEN_PEER_IP] = peer_ip_open

        # Parse version
        if FTL_ATTR_BGP_OPEN_VERSION >= 0:
            open_record[FTL_ATTR_BGP_OPEN_VERSION] = msg_bytes[offset]
        offset += 1

        # Parse my AS (2 byte only)
        if FTL_ATTR_BGP_OPEN_MY_AS >= 0:
            open_record[FTL_ATTR_BGP_OPEN_MY_AS] = struct_unpack(STRUCT_2B, msg_bytes[offset:offset + 2])[0]
        offset += 2

        # Parse hold time
        if FTL_ATTR_BGP_OPEN_HOLD_TIME >= 0:
            open_record[FTL_ATTR_BGP_OPEN_HOLD_TIME] = struct_unpack(STRUCT_2B, msg_bytes[offset:offset + 2])[0]
        offset += 2

        # Parse BGP ID (IPv4 only)
        if FTL_ATTR_BGP_OPEN_BGP_ID >= 0:
            bgp_id = msg_bytes[offset:offset + 4]
            if FTL_ATTR_BGP_OPEN_BGP_ID_HUMAN:
                open_record[FTL_ATTR_BGP_OPEN_BGP_ID] = socket_inet_ntop(AF_INET, bgp_id)
            else:
                open_record[FTL_ATTR_BGP_OPEN_BGP_ID] = struct_unpack(STRUCT_4B, bgp_id)[0]
        offset += 4

        ###################
        # BGP OPEN PARAMS #
        ###################

        # Parse optional param length
        optlen = msg_bytes[offset]
        offset += 1

        # Parse optional params
        cur_offset, end_offset = offset, offset + optlen
        offset = end_offset
        while cur_offset < end_offset:

            # Parse optional parameter type
            otype = msg_bytes[cur_offset]
            cur_offset += 1

            # Parse optional parameter length
            olen = msg_bytes[cur_offset]
            cur_offset += 1

            # Prepare optional parameter byte offsets
            c_offset, e_offset = cur_offset, cur_offset + olen
            cur_offset = e_offset

            # ---------------------------------------------------------------
            # [RFC3392] 4. Capabilities Optional Parameter (Parameter Type 2)
            # ---------------------------------------------------------------
            # +------------------------------+
            # | Capability Code (1 octet)    |
            # +------------------------------+
            # | Capability Length (1 octet)  |
            # +------------------------------+
            # | Capability Value (variable)  |
            # +------------------------------+

            # Parse capabilities optional parameter
            if otype == BGP_PARAMS_CAPABILITIES:
                if FTL_ATTR_BGP_OPEN_CAPABILITIES >= 0:

                    # Parse capabilities
                    while c_offset < e_offset:

                        # Parse capability code
                        ccode = msg_bytes[c_offset]
                        ccode_stats = ccode
                        if FTL_ATTR_BGP_OPEN_CAPABILITIES_HUMAN:
                            ccode = FTL_ATTR_BGP_OPEN_CAPABILITY_TO_STR.get(ccode, str(ccode))
                        c_offset += 1

                        # Parse capability length
                        clen = msg_bytes[c_offset]
                        c_offset += 1

                        # Parse capability data
                        cdata = None
                        if clen > 0:
                            cdata = msg_bytes[c_offset:c_offset + clen]
                            if FTL_ATTR_BGP_OPEN_CAPABILITIES_HUMAN:
                                cdata = ' '.join('{:02x}'.format(byte) for byte in cdata)
                            else:
                                cdata = base64.b64encode(cdata).decode('ascii')
                            c_offset += clen

                        # Add capability code and data
                        if open_record[FTL_ATTR_BGP_OPEN_CAPABILITIES] is None:
                            open_record[FTL_ATTR_BGP_OPEN_CAPABILITIES] = tuple([(ccode, cdata)])
                        else:
                            open_record[FTL_ATTR_BGP_OPEN_CAPABILITIES] += tuple([(ccode, cdata)])

                        # Update stats record
                        if FTL_RECORD_BGP_STATS:

                            # Add BGP message type
                            if FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES >= 0:
                                if FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES_HUMAN:
                                    ccode_stats = FTL_ATTR_BGP_OPEN_CAPABILITY_TO_STR.get(ccode_stats, ccode_stats)
                                ccode_stats = str(ccode_stats)
                                stats_record_mrt_bgp_cap = stats_record[FTL_ATTR_BGP_STATS_MRT_BGP_CAPABILITY_TYPES]
                                stats_record_mrt_bgp_cap[ccode_stats] = stats_record_mrt_bgp_cap.get(ccode_stats, 0) + 1

            # Parse unsupported optional parameters
            elif FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN >= 0:  # pylint: disable=confusing-consecutive-elif
                odata = msg_bytes[c_offset:e_offset]
                if FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN_HUMAN:
                    odata = ' '.join('{:02x}'.format(byte) for byte in odata)
                else:
                    odata = base64.b64encode(odata).decode('ascii')
                if open_record[FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN] is None:
                    open_record[FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN] = tuple([(otype, odata)])
                else:
                    open_record[FTL_ATTR_BGP_OPEN_PARAMS_UNKNOWN] += tuple([(otype, odata)])

        # Yield final open record
        if FTL_RECORD_BGP_OPEN:
            yield open_emit(open_record)
